# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import commit_pb2 as commit__pb2


class CommitServiceStub(object):
    """CommitService is a service which provides RPCs that interact with Git
    commits.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListCommits = channel.unary_stream(
                '/gitaly.CommitService/ListCommits',
                request_serializer=commit__pb2.ListCommitsRequest.SerializeToString,
                response_deserializer=commit__pb2.ListCommitsResponse.FromString,
                )
        self.ListAllCommits = channel.unary_stream(
                '/gitaly.CommitService/ListAllCommits',
                request_serializer=commit__pb2.ListAllCommitsRequest.SerializeToString,
                response_deserializer=commit__pb2.ListAllCommitsResponse.FromString,
                )
        self.CommitIsAncestor = channel.unary_unary(
                '/gitaly.CommitService/CommitIsAncestor',
                request_serializer=commit__pb2.CommitIsAncestorRequest.SerializeToString,
                response_deserializer=commit__pb2.CommitIsAncestorResponse.FromString,
                )
        self.TreeEntry = channel.unary_stream(
                '/gitaly.CommitService/TreeEntry',
                request_serializer=commit__pb2.TreeEntryRequest.SerializeToString,
                response_deserializer=commit__pb2.TreeEntryResponse.FromString,
                )
        self.CountCommits = channel.unary_unary(
                '/gitaly.CommitService/CountCommits',
                request_serializer=commit__pb2.CountCommitsRequest.SerializeToString,
                response_deserializer=commit__pb2.CountCommitsResponse.FromString,
                )
        self.CountDivergingCommits = channel.unary_unary(
                '/gitaly.CommitService/CountDivergingCommits',
                request_serializer=commit__pb2.CountDivergingCommitsRequest.SerializeToString,
                response_deserializer=commit__pb2.CountDivergingCommitsResponse.FromString,
                )
        self.GetTreeEntries = channel.unary_stream(
                '/gitaly.CommitService/GetTreeEntries',
                request_serializer=commit__pb2.GetTreeEntriesRequest.SerializeToString,
                response_deserializer=commit__pb2.GetTreeEntriesResponse.FromString,
                )
        self.ListFiles = channel.unary_stream(
                '/gitaly.CommitService/ListFiles',
                request_serializer=commit__pb2.ListFilesRequest.SerializeToString,
                response_deserializer=commit__pb2.ListFilesResponse.FromString,
                )
        self.FindCommit = channel.unary_unary(
                '/gitaly.CommitService/FindCommit',
                request_serializer=commit__pb2.FindCommitRequest.SerializeToString,
                response_deserializer=commit__pb2.FindCommitResponse.FromString,
                )
        self.CommitStats = channel.unary_unary(
                '/gitaly.CommitService/CommitStats',
                request_serializer=commit__pb2.CommitStatsRequest.SerializeToString,
                response_deserializer=commit__pb2.CommitStatsResponse.FromString,
                )
        self.FindAllCommits = channel.unary_stream(
                '/gitaly.CommitService/FindAllCommits',
                request_serializer=commit__pb2.FindAllCommitsRequest.SerializeToString,
                response_deserializer=commit__pb2.FindAllCommitsResponse.FromString,
                )
        self.FindCommits = channel.unary_stream(
                '/gitaly.CommitService/FindCommits',
                request_serializer=commit__pb2.FindCommitsRequest.SerializeToString,
                response_deserializer=commit__pb2.FindCommitsResponse.FromString,
                )
        self.CommitLanguages = channel.unary_unary(
                '/gitaly.CommitService/CommitLanguages',
                request_serializer=commit__pb2.CommitLanguagesRequest.SerializeToString,
                response_deserializer=commit__pb2.CommitLanguagesResponse.FromString,
                )
        self.RawBlame = channel.unary_stream(
                '/gitaly.CommitService/RawBlame',
                request_serializer=commit__pb2.RawBlameRequest.SerializeToString,
                response_deserializer=commit__pb2.RawBlameResponse.FromString,
                )
        self.LastCommitForPath = channel.unary_unary(
                '/gitaly.CommitService/LastCommitForPath',
                request_serializer=commit__pb2.LastCommitForPathRequest.SerializeToString,
                response_deserializer=commit__pb2.LastCommitForPathResponse.FromString,
                )
        self.ListLastCommitsForTree = channel.unary_stream(
                '/gitaly.CommitService/ListLastCommitsForTree',
                request_serializer=commit__pb2.ListLastCommitsForTreeRequest.SerializeToString,
                response_deserializer=commit__pb2.ListLastCommitsForTreeResponse.FromString,
                )
        self.CommitsByMessage = channel.unary_stream(
                '/gitaly.CommitService/CommitsByMessage',
                request_serializer=commit__pb2.CommitsByMessageRequest.SerializeToString,
                response_deserializer=commit__pb2.CommitsByMessageResponse.FromString,
                )
        self.ListCommitsByOid = channel.unary_stream(
                '/gitaly.CommitService/ListCommitsByOid',
                request_serializer=commit__pb2.ListCommitsByOidRequest.SerializeToString,
                response_deserializer=commit__pb2.ListCommitsByOidResponse.FromString,
                )
        self.ListCommitsByRefName = channel.unary_stream(
                '/gitaly.CommitService/ListCommitsByRefName',
                request_serializer=commit__pb2.ListCommitsByRefNameRequest.SerializeToString,
                response_deserializer=commit__pb2.ListCommitsByRefNameResponse.FromString,
                )
        self.FilterShasWithSignatures = channel.stream_stream(
                '/gitaly.CommitService/FilterShasWithSignatures',
                request_serializer=commit__pb2.FilterShasWithSignaturesRequest.SerializeToString,
                response_deserializer=commit__pb2.FilterShasWithSignaturesResponse.FromString,
                )
        self.GetCommitSignatures = channel.unary_stream(
                '/gitaly.CommitService/GetCommitSignatures',
                request_serializer=commit__pb2.GetCommitSignaturesRequest.SerializeToString,
                response_deserializer=commit__pb2.GetCommitSignaturesResponse.FromString,
                )
        self.GetCommitMessages = channel.unary_stream(
                '/gitaly.CommitService/GetCommitMessages',
                request_serializer=commit__pb2.GetCommitMessagesRequest.SerializeToString,
                response_deserializer=commit__pb2.GetCommitMessagesResponse.FromString,
                )
        self.CheckObjectsExist = channel.stream_stream(
                '/gitaly.CommitService/CheckObjectsExist',
                request_serializer=commit__pb2.CheckObjectsExistRequest.SerializeToString,
                response_deserializer=commit__pb2.CheckObjectsExistResponse.FromString,
                )


class CommitServiceServicer(object):
    """CommitService is a service which provides RPCs that interact with Git
    commits.
    """

    def ListCommits(self, request, context):
        """ListCommits lists all commits reachable via a set of references by doing a
        graph walk. This deprecates FindAllCommits and FindCommits (except Follow
        is not yet supported). Any unknown revisions will cause the RPC to fail.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListAllCommits(self, request, context):
        """ListAllCommits lists all commits present in the repository, including
        those not reachable by any reference.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CommitIsAncestor(self, request, context):
        """CommitIsAncestor checks whether a provided commit is the ancestor of
        another commit.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TreeEntry(self, request, context):
        """TreeEntry provides the tree entry for the provided path and revision. The data
        is broken into chunks and streamed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CountCommits(self, request, context):
        """CountCommits provides the number of commits which adhere to the given filters.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CountDivergingCommits(self, request, context):
        """CountDivergingCommits provides the number of diverging commits between two revisions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTreeEntries(self, request, context):
        """GetTreeEntries provides the tree entries for the provided path and revision. This includes
        subtrees present under the tree with the option of recursive fetching.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListFiles(self, request, context):
        """ListFiles lists all the files (including files in sub-dirs) present in the working tree
        of a given treeish.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindCommit(self, request, context):
        """FindCommit finds a commit for a given commitish. Returns nil if the commit is not found.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CommitStats(self, request, context):
        """CommitStats provides the stats for a given commitish.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindAllCommits(self, request, context):
        """FindAllCommits lists all the commits which can be traversed from the
        provided commitish.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindCommits(self, request, context):
        """FindCommits lists all the commits which are associated with the provided revision
        and paths.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CommitLanguages(self, request, context):
        """CommitLanguages detects the source code languages of the whole tree for a
        given commit. Returns an error in case no languages could be detected.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RawBlame(self, request, context):
        """RawBlame blames lines in a blob to when they have last been changed. Returns the raw output of the git-blame(1)
        command.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LastCommitForPath(self, request, context):
        """LastCommitForPath returns the last commit that has changed a given path.

        The following special cases apply and have grown historically:

        - Absolute paths that or relative paths that escape the repository root will cause an error.
        - A nonexistent path inside the repository leads to a successful but empty response.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListLastCommitsForTree(self, request, context):
        """ListLastCommitsForTree lists the last commits for a given tree.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CommitsByMessage(self, request, context):
        """CommitsByMessage list commits which match the provided query.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListCommitsByOid(self, request, context):
        """ListCommitsByOid lists the commits for the provided commitish object IDs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListCommitsByRefName(self, request, context):
        """ListCommitsByRefName lists the commits for the provided references.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FilterShasWithSignatures(self, request_iterator, context):
        """FilterShasWithSignatures filters out signed commits.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCommitSignatures(self, request, context):
        """GetCommitSignatures parses the commit signature information for the provided commitish object IDs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCommitMessages(self, request, context):
        """GetCommitMessages returns the commits messages for the provided commitish object IDs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CheckObjectsExist(self, request_iterator, context):
        """CheckObjectsExist will check for the existence of revisions against a
        repository. It returns two sets of data. An array containing the revisions
        from the input that it found on the repository, and an array that contains all
        revisions from the input it did not find on the repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CommitServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ListCommits': grpc.unary_stream_rpc_method_handler(
                    servicer.ListCommits,
                    request_deserializer=commit__pb2.ListCommitsRequest.FromString,
                    response_serializer=commit__pb2.ListCommitsResponse.SerializeToString,
            ),
            'ListAllCommits': grpc.unary_stream_rpc_method_handler(
                    servicer.ListAllCommits,
                    request_deserializer=commit__pb2.ListAllCommitsRequest.FromString,
                    response_serializer=commit__pb2.ListAllCommitsResponse.SerializeToString,
            ),
            'CommitIsAncestor': grpc.unary_unary_rpc_method_handler(
                    servicer.CommitIsAncestor,
                    request_deserializer=commit__pb2.CommitIsAncestorRequest.FromString,
                    response_serializer=commit__pb2.CommitIsAncestorResponse.SerializeToString,
            ),
            'TreeEntry': grpc.unary_stream_rpc_method_handler(
                    servicer.TreeEntry,
                    request_deserializer=commit__pb2.TreeEntryRequest.FromString,
                    response_serializer=commit__pb2.TreeEntryResponse.SerializeToString,
            ),
            'CountCommits': grpc.unary_unary_rpc_method_handler(
                    servicer.CountCommits,
                    request_deserializer=commit__pb2.CountCommitsRequest.FromString,
                    response_serializer=commit__pb2.CountCommitsResponse.SerializeToString,
            ),
            'CountDivergingCommits': grpc.unary_unary_rpc_method_handler(
                    servicer.CountDivergingCommits,
                    request_deserializer=commit__pb2.CountDivergingCommitsRequest.FromString,
                    response_serializer=commit__pb2.CountDivergingCommitsResponse.SerializeToString,
            ),
            'GetTreeEntries': grpc.unary_stream_rpc_method_handler(
                    servicer.GetTreeEntries,
                    request_deserializer=commit__pb2.GetTreeEntriesRequest.FromString,
                    response_serializer=commit__pb2.GetTreeEntriesResponse.SerializeToString,
            ),
            'ListFiles': grpc.unary_stream_rpc_method_handler(
                    servicer.ListFiles,
                    request_deserializer=commit__pb2.ListFilesRequest.FromString,
                    response_serializer=commit__pb2.ListFilesResponse.SerializeToString,
            ),
            'FindCommit': grpc.unary_unary_rpc_method_handler(
                    servicer.FindCommit,
                    request_deserializer=commit__pb2.FindCommitRequest.FromString,
                    response_serializer=commit__pb2.FindCommitResponse.SerializeToString,
            ),
            'CommitStats': grpc.unary_unary_rpc_method_handler(
                    servicer.CommitStats,
                    request_deserializer=commit__pb2.CommitStatsRequest.FromString,
                    response_serializer=commit__pb2.CommitStatsResponse.SerializeToString,
            ),
            'FindAllCommits': grpc.unary_stream_rpc_method_handler(
                    servicer.FindAllCommits,
                    request_deserializer=commit__pb2.FindAllCommitsRequest.FromString,
                    response_serializer=commit__pb2.FindAllCommitsResponse.SerializeToString,
            ),
            'FindCommits': grpc.unary_stream_rpc_method_handler(
                    servicer.FindCommits,
                    request_deserializer=commit__pb2.FindCommitsRequest.FromString,
                    response_serializer=commit__pb2.FindCommitsResponse.SerializeToString,
            ),
            'CommitLanguages': grpc.unary_unary_rpc_method_handler(
                    servicer.CommitLanguages,
                    request_deserializer=commit__pb2.CommitLanguagesRequest.FromString,
                    response_serializer=commit__pb2.CommitLanguagesResponse.SerializeToString,
            ),
            'RawBlame': grpc.unary_stream_rpc_method_handler(
                    servicer.RawBlame,
                    request_deserializer=commit__pb2.RawBlameRequest.FromString,
                    response_serializer=commit__pb2.RawBlameResponse.SerializeToString,
            ),
            'LastCommitForPath': grpc.unary_unary_rpc_method_handler(
                    servicer.LastCommitForPath,
                    request_deserializer=commit__pb2.LastCommitForPathRequest.FromString,
                    response_serializer=commit__pb2.LastCommitForPathResponse.SerializeToString,
            ),
            'ListLastCommitsForTree': grpc.unary_stream_rpc_method_handler(
                    servicer.ListLastCommitsForTree,
                    request_deserializer=commit__pb2.ListLastCommitsForTreeRequest.FromString,
                    response_serializer=commit__pb2.ListLastCommitsForTreeResponse.SerializeToString,
            ),
            'CommitsByMessage': grpc.unary_stream_rpc_method_handler(
                    servicer.CommitsByMessage,
                    request_deserializer=commit__pb2.CommitsByMessageRequest.FromString,
                    response_serializer=commit__pb2.CommitsByMessageResponse.SerializeToString,
            ),
            'ListCommitsByOid': grpc.unary_stream_rpc_method_handler(
                    servicer.ListCommitsByOid,
                    request_deserializer=commit__pb2.ListCommitsByOidRequest.FromString,
                    response_serializer=commit__pb2.ListCommitsByOidResponse.SerializeToString,
            ),
            'ListCommitsByRefName': grpc.unary_stream_rpc_method_handler(
                    servicer.ListCommitsByRefName,
                    request_deserializer=commit__pb2.ListCommitsByRefNameRequest.FromString,
                    response_serializer=commit__pb2.ListCommitsByRefNameResponse.SerializeToString,
            ),
            'FilterShasWithSignatures': grpc.stream_stream_rpc_method_handler(
                    servicer.FilterShasWithSignatures,
                    request_deserializer=commit__pb2.FilterShasWithSignaturesRequest.FromString,
                    response_serializer=commit__pb2.FilterShasWithSignaturesResponse.SerializeToString,
            ),
            'GetCommitSignatures': grpc.unary_stream_rpc_method_handler(
                    servicer.GetCommitSignatures,
                    request_deserializer=commit__pb2.GetCommitSignaturesRequest.FromString,
                    response_serializer=commit__pb2.GetCommitSignaturesResponse.SerializeToString,
            ),
            'GetCommitMessages': grpc.unary_stream_rpc_method_handler(
                    servicer.GetCommitMessages,
                    request_deserializer=commit__pb2.GetCommitMessagesRequest.FromString,
                    response_serializer=commit__pb2.GetCommitMessagesResponse.SerializeToString,
            ),
            'CheckObjectsExist': grpc.stream_stream_rpc_method_handler(
                    servicer.CheckObjectsExist,
                    request_deserializer=commit__pb2.CheckObjectsExistRequest.FromString,
                    response_serializer=commit__pb2.CheckObjectsExistResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gitaly.CommitService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class CommitService(object):
    """CommitService is a service which provides RPCs that interact with Git
    commits.
    """

    @staticmethod
    def ListCommits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/ListCommits',
            commit__pb2.ListCommitsRequest.SerializeToString,
            commit__pb2.ListCommitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListAllCommits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/ListAllCommits',
            commit__pb2.ListAllCommitsRequest.SerializeToString,
            commit__pb2.ListAllCommitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CommitIsAncestor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.CommitService/CommitIsAncestor',
            commit__pb2.CommitIsAncestorRequest.SerializeToString,
            commit__pb2.CommitIsAncestorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TreeEntry(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/TreeEntry',
            commit__pb2.TreeEntryRequest.SerializeToString,
            commit__pb2.TreeEntryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CountCommits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.CommitService/CountCommits',
            commit__pb2.CountCommitsRequest.SerializeToString,
            commit__pb2.CountCommitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CountDivergingCommits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.CommitService/CountDivergingCommits',
            commit__pb2.CountDivergingCommitsRequest.SerializeToString,
            commit__pb2.CountDivergingCommitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTreeEntries(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/GetTreeEntries',
            commit__pb2.GetTreeEntriesRequest.SerializeToString,
            commit__pb2.GetTreeEntriesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListFiles(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/ListFiles',
            commit__pb2.ListFilesRequest.SerializeToString,
            commit__pb2.ListFilesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FindCommit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.CommitService/FindCommit',
            commit__pb2.FindCommitRequest.SerializeToString,
            commit__pb2.FindCommitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CommitStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.CommitService/CommitStats',
            commit__pb2.CommitStatsRequest.SerializeToString,
            commit__pb2.CommitStatsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FindAllCommits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/FindAllCommits',
            commit__pb2.FindAllCommitsRequest.SerializeToString,
            commit__pb2.FindAllCommitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FindCommits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/FindCommits',
            commit__pb2.FindCommitsRequest.SerializeToString,
            commit__pb2.FindCommitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CommitLanguages(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.CommitService/CommitLanguages',
            commit__pb2.CommitLanguagesRequest.SerializeToString,
            commit__pb2.CommitLanguagesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RawBlame(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/RawBlame',
            commit__pb2.RawBlameRequest.SerializeToString,
            commit__pb2.RawBlameResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LastCommitForPath(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.CommitService/LastCommitForPath',
            commit__pb2.LastCommitForPathRequest.SerializeToString,
            commit__pb2.LastCommitForPathResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListLastCommitsForTree(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/ListLastCommitsForTree',
            commit__pb2.ListLastCommitsForTreeRequest.SerializeToString,
            commit__pb2.ListLastCommitsForTreeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CommitsByMessage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/CommitsByMessage',
            commit__pb2.CommitsByMessageRequest.SerializeToString,
            commit__pb2.CommitsByMessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListCommitsByOid(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/ListCommitsByOid',
            commit__pb2.ListCommitsByOidRequest.SerializeToString,
            commit__pb2.ListCommitsByOidResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListCommitsByRefName(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/ListCommitsByRefName',
            commit__pb2.ListCommitsByRefNameRequest.SerializeToString,
            commit__pb2.ListCommitsByRefNameResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FilterShasWithSignatures(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/gitaly.CommitService/FilterShasWithSignatures',
            commit__pb2.FilterShasWithSignaturesRequest.SerializeToString,
            commit__pb2.FilterShasWithSignaturesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetCommitSignatures(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/GetCommitSignatures',
            commit__pb2.GetCommitSignaturesRequest.SerializeToString,
            commit__pb2.GetCommitSignaturesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetCommitMessages(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.CommitService/GetCommitMessages',
            commit__pb2.GetCommitMessagesRequest.SerializeToString,
            commit__pb2.GetCommitMessagesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CheckObjectsExist(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/gitaly.CommitService/CheckObjectsExist',
            commit__pb2.CheckObjectsExistRequest.SerializeToString,
            commit__pb2.CheckObjectsExistResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: server.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from . import lint_pb2 as lint__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cserver.proto\x12\x06gitaly\x1a\x1egoogle/protobuf/duration.proto\x1a\nlint.proto\"\x13\n\x11ServerInfoRequest\"\x96\x02\n\x12ServerInfoResponse\x12\x16\n\x0eserver_version\x18\x01 \x01(\t\x12\x13\n\x0bgit_version\x18\x02 \x01(\t\x12\x42\n\x10storage_statuses\x18\x03 \x03(\x0b\x32(.gitaly.ServerInfoResponse.StorageStatus\x1a\x8e\x01\n\rStorageStatus\x12\x14\n\x0cstorage_name\x18\x01 \x01(\t\x12\x10\n\x08readable\x18\x02 \x01(\x08\x12\x11\n\twriteable\x18\x03 \x01(\x08\x12\x0f\n\x07\x66s_type\x18\x04 \x01(\t\x12\x15\n\rfilesystem_id\x18\x05 \x01(\t\x12\x1a\n\x12replication_factor\x18\x06 \x01(\r\"\x17\n\x15\x44iskStatisticsRequest\"\xa8\x01\n\x16\x44iskStatisticsResponse\x12\x46\n\x10storage_statuses\x18\x01 \x03(\x0b\x32,.gitaly.DiskStatisticsResponse.StorageStatus\x1a\x46\n\rStorageStatus\x12\x14\n\x0cstorage_name\x18\x01 \x01(\t\x12\x11\n\tavailable\x18\x02 \x01(\x03\x12\x0c\n\x04used\x18\x03 \x01(\x03\"x\n\x12\x43lockSyncedRequest\x12\x10\n\x08ntp_host\x18\x01 \x01(\t\x12\x32\n\x0f\x64rift_threshold\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationJ\x04\x08\x02\x10\x03R\x16\x64rift_threshold_millis\"%\n\x13\x43lockSyncedResponse\x12\x0e\n\x06synced\x18\x01 \x01(\x08\"C\n\x15ReadinessCheckRequest\x12*\n\x07timeout\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"\xab\x02\n\x16ReadinessCheckResponse\x12\x38\n\x0bok_response\x18\x01 \x01(\x0b\x32!.gitaly.ReadinessCheckResponse.OkH\x00\x12\x42\n\x10\x66\x61ilure_response\x18\x02 \x01(\x0b\x32&.gitaly.ReadinessCheckResponse.FailureH\x00\x1a\x04\n\x02Ok\x1a\x82\x01\n\x07\x46\x61ilure\x12\x46\n\rfailed_checks\x18\x01 \x03(\x0b\x32/.gitaly.ReadinessCheckResponse.Failure.Response\x1a/\n\x08Response\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rerror_message\x18\x02 \x01(\tB\x08\n\x06Result2\xc4\x02\n\rServerService\x12\x43\n\nServerInfo\x12\x19.gitaly.ServerInfoRequest\x1a\x1a.gitaly.ServerInfoResponse\x12O\n\x0e\x44iskStatistics\x12\x1d.gitaly.DiskStatisticsRequest\x1a\x1e.gitaly.DiskStatisticsResponse\x12\x46\n\x0b\x43lockSynced\x12\x1a.gitaly.ClockSyncedRequest\x1a\x1b.gitaly.ClockSyncedResponse\x12O\n\x0eReadinessCheck\x12\x1d.gitaly.ReadinessCheckRequest\x1a\x1e.gitaly.ReadinessCheckResponse\x1a\x04\xf0\x97(\x01\x42\x34Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'server_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypb'
  _SERVERSERVICE._options = None
  _SERVERSERVICE._serialized_options = b'\360\227(\001'
  _globals['_SERVERINFOREQUEST']._serialized_start=68
  _globals['_SERVERINFOREQUEST']._serialized_end=87
  _globals['_SERVERINFORESPONSE']._serialized_start=90
  _globals['_SERVERINFORESPONSE']._serialized_end=368
  _globals['_SERVERINFORESPONSE_STORAGESTATUS']._serialized_start=226
  _globals['_SERVERINFORESPONSE_STORAGESTATUS']._serialized_end=368
  _globals['_DISKSTATISTICSREQUEST']._serialized_start=370
  _globals['_DISKSTATISTICSREQUEST']._serialized_end=393
  _globals['_DISKSTATISTICSRESPONSE']._serialized_start=396
  _globals['_DISKSTATISTICSRESPONSE']._serialized_end=564
  _globals['_DISKSTATISTICSRESPONSE_STORAGESTATUS']._serialized_start=494
  _globals['_DISKSTATISTICSRESPONSE_STORAGESTATUS']._serialized_end=564
  _globals['_CLOCKSYNCEDREQUEST']._serialized_start=566
  _globals['_CLOCKSYNCEDREQUEST']._serialized_end=686
  _globals['_CLOCKSYNCEDRESPONSE']._serialized_start=688
  _globals['_CLOCKSYNCEDRESPONSE']._serialized_end=725
  _globals['_READINESSCHECKREQUEST']._serialized_start=727
  _globals['_READINESSCHECKREQUEST']._serialized_end=794
  _globals['_READINESSCHECKRESPONSE']._serialized_start=797
  _globals['_READINESSCHECKRESPONSE']._serialized_end=1096
  _globals['_READINESSCHECKRESPONSE_OK']._serialized_start=949
  _globals['_READINESSCHECKRESPONSE_OK']._serialized_end=953
  _globals['_READINESSCHECKRESPONSE_FAILURE']._serialized_start=956
  _globals['_READINESSCHECKRESPONSE_FAILURE']._serialized_end=1086
  _globals['_READINESSCHECKRESPONSE_FAILURE_RESPONSE']._serialized_start=1039
  _globals['_READINESSCHECKRESPONSE_FAILURE_RESPONSE']._serialized_end=1086
  _globals['_SERVERSERVICE']._serialized_start=1099
  _globals['_SERVERSERVICE']._serialized_end=1423
# @@protoc_insertion_point(module_scope)



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

# TODO bring full coverage directly from here, because there might
# be points in time where it won't be used at all by gRPC methods
import pytest

from ..testing.context import FakeServicerContext
from .. import feature


class FakeContext(FakeServicerContext):

    def __init__(self, invocation_metadata):
        self._invocation_metadata = invocation_metadata

    def invocation_metadata(self):
        return self._invocation_metadata


@pytest.fixture
def feature_flags(monkeypatch):
    monkeypatch.setattr(feature, 'FEATURE_FLAGS',
                        {'default-disabled': False,
                         'default-enabled': True,
                         })
    # in case we do a full tests run with the all-enabling environment
    # variable (can be useful to detect flags that we should implement)
    monkeypatch.setattr(feature, 'ALL_ENABLED', False)
    yield monkeypatch


def test_is_enabled_defaults(feature_flags):
    context = FakeContext(())
    assert feature.is_enabled(context, 'default-enabled')
    assert not feature.is_enabled(context, 'default-disabled')
    with pytest.raises(feature.UndefinedError) as exc_info:
        feature.is_enabled(context, 'not-defined')
    assert exc_info.value.args == ('not-defined', )


def test_is_enabled_all_enabled(feature_flags):
    context = FakeContext(())
    feature_flags.setattr(feature, 'ALL_ENABLED', True)
    assert feature.is_enabled(context, 'default-disabled')

    # still raises when HGitaly wants to use an undefined flag
    with pytest.raises(feature.UndefinedError) as exc_info:
        feature.is_enabled(context, 'not-defined')
    assert exc_info.value.args == ('not-defined', )


def test_is_enabled_context(feature_flags):
    context = FakeContext((
        ('gitaly-feature-default-disabled', 'true'),
    ))
    assert feature.is_enabled(context, 'default-disabled')


def test_as_grpc_metadata():
    assert feature.as_grpc_metadata(None) is None
    assert feature.as_grpc_metadata((
        ('my-flag', True),
        ('their-flag', False),
    )) == [
        (b'gitaly-feature-my-flag', b'true'),
        (b'gitaly-feature-their-flag', b'false'),
    ]



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import attr
import logging
import psutil
from multiprocessing import Process

logger = logging.getLogger(__name__)


@attr.s
class WorkerProcess:
    """Wrapping a :class:`Process` instance with more utilities."""

    process_callable = attr.ib()
    """The Python callable to run in the process."""

    process_args = attr.ib()
    """The arguments to pass to the underlying Python callable."""

    process = attr.ib(default=None)
    """The :class:`Process` instance."""

    monitoring = attr.ib(default=None)
    """The :class:`psutil.Process` used for monitoring."""

    restart_done_workers = attr.ib(default=True)
    """Control whether :meth:`watch` should restart done workers.

    Not restarting workers is important for tests, where we have short lived
    workers.
    """

    max_rss = attr.ib(default=1 << 30)
    """Maximum memory Resident Set Size before :meth:`watch` restarts.

    Default is 1GiB.
    """

    graceful_shutdown_timeout_seconds = attr.ib(default=300)
    """Maximum time alocated to worker to handle current request upon SIGTERM.
    """

    worker_kwargs_attrs = ('graceful_shutdown_timeout_seconds',
                           )
    """Attributes to forward as worker process keyword arguments"""

    @classmethod
    def run(cls, *a, **kw):
        wp = cls(*a, **kw)
        wp.init_process()
        wp.start()
        return wp

    def __str__(self):
        pid = 'no process' if self.process is None else self.process.pid
        return f"Worker {self.process_args[0]} (pid: {pid})"

    def init_process(self):
        self.process = Process(target=self.process_callable,
                               args=self.process_args,
                               kwargs={a: getattr(self, a)
                                       for a in self.worker_kwargs_attrs})

    def start(self):
        self.process.start()
        self.monitoring = psutil.Process(self.process.pid)

    def join(self):
        if self.process is not None:
            self.process.join()

    def is_alive(self):
        return False if self.process is None else self.process.is_alive()

    def terminate(self):
        if self.process is not None:
            self.process.terminate()

    def restart(self):
        logger.info("Restarting %s", self)
        self.terminate()
        self.join()
        self.init_process()
        self.start()

    def watch(self):
        """Checks process statistics and restart it if needed.

        :return: ``True`` if monitoring should go on, ``False``
           otherwise.
        """
        if self.monitoring is None:
            logger.debug("%s process not started", self)
            return True

        try:
            restart_or_reap = self.monitoring.status() == 'zombie'
            mem_stats = self.monitoring.memory_info()
        except psutil.NoSuchProcess as exc:
            logger.warning("%s %s", self, exc)
            restart_or_reap = True

        if restart_or_reap:
            if self.restart_done_workers:
                self.restart()
                return True
            self.join()
            return False

        rss = mem_stats.rss
        over_limit_restart = rss > self.max_rss

        log = logger.warning if over_limit_restart else logger.info
        log("%s %sRSS=%dMiB. Full mem stats: %r",
            self, "over limit, will restart: " if over_limit_restart else '',
            rss >> 20, mem_stats)
        if over_limit_restart:
            self.restart()
        return True



# Copyright 2024 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest
import random
import time

from grpc import RpcError, StatusCode
from psutil import Process

from heptapod.testhelpers.hg import RepoWrapper

from hgitaly.stub.mercurial_repository_pb2 import (
    MercurialPeer,
    PullRequest,
)
from hgitaly.stub.repository_pb2 import (
    CreateRepositoryRequest,
)
from hgitaly.stub.shared_pb2 import (
    Repository,
)

from hgitaly.stub.mercurial_repository_pb2_grpc import (
    MercurialRepositoryServiceStub,
)
from hgitaly.stub.repository_pb2_grpc import (
    RepositoryServiceStub,
)

from .rhgitaly import RHGitalyServer

from . import skip_comparison_tests
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip


parametrize = pytest.mark.parametrize


def test_rhgitaly_pull(rhgitaly, server_repos_root):
    fixture = rhgitaly
    repo_wrapper = fixture.hg_repo_wrapper
    rpc_helper = fixture.rpc_helper(stub_cls=MercurialRepositoryServiceStub,
                                    method_name='Pull',
                                    request_cls=PullRequest)

    src_repo_name = 'src-%s.hg' % hex(random.getrandbits(64))[2:]
    src_repo = RepoWrapper.init(server_repos_root / src_repo_name)
    src_hex = src_repo.commit_file('foo', message='foo-pull').hex()

    peer = MercurialPeer(url=str(src_repo.path))
    resp = rpc_helper.rpc('rhgitaly', remote_peer=peer)
    assert resp.new_changesets

    fixture.invalidate()
    assert repo_wrapper.repo[src_hex].description() == b'foo-pull'

    resp = rpc_helper.rpc('rhgitaly', remote_peer=peer)
    assert not resp.new_changesets


@parametrize('cause', ['client', 'server-shutdown'])
def test_pull_rhgitaly_termination(grpc_channel, server_repos_root, cause):
    # Here we are using the regular HGitaly grpc channel to create the repo,
    hgitaly_service = RepositoryServiceStub(grpc_channel)
    repo_path = 'repo-%s.hg' % hex(random.getrandbits(64))[2:]
    gl_repo = Repository(relative_path=repo_path, storage_name="default")
    hgitaly_service.CreateRepository(
        CreateRepositoryRequest(repository=gl_repo))

    sleeper_path = server_repos_root / 'sleeper'
    with open(sleeper_path, 'w') as sf:
        sf.writelines(("#!/bin/sh\n",
                       "echo $@\n",
                       "sleep 120\n"))

    sleeper_path.chmod(0o755)

    rhgitaly = RHGitalyServer(
        server_repos_root,
        socket_name='rhgitaly-oneshot.socket',
        sidecar_address='http://sidecar.example',  # no attempt to open channel
        env_overrides={'RHGITALY_HG_EXECUTABLE': str(sleeper_path)},
    )
    with rhgitaly.running() as rhgitaly_channel:
        mercurial_repo_service = MercurialRepositoryServiceStub(
            rhgitaly_channel)
        peer = MercurialPeer(url='/does/not/matter')
        async_meth = mercurial_repo_service.Pull.future
        future = async_meth(PullRequest(remote_peer=peer, repository=gl_repo))
        start = time.time()
        while time.time() - start < 1:
            rhgitaly_children = Process(rhgitaly.pid).children()
            sleepers = [p for p in rhgitaly_children if p.name() == 'sleeper']
            if sleepers:
                break
            time.sleep(0.01)

        sleeper = sleepers[0]

        if cause == 'client':
            future.cancel()
            # client lib will forbid us to call for the result, so we cannot
            # check that the server sends a `CANCELLED` status code.
            # But our major point is the termination of the subprocess

    if cause == 'server-shutdown':
        with pytest.raises(RpcError) as exc_info:
            future.result()
        assert exc_info.value.code() == StatusCode.CANCELLED
        assert 'Channel closed' in exc_info.value.details()

    # according to documentation at
    #   https://psutil.readthedocs.io/en/latest/#psutil.Process.is_running,
    # this does not suffer from PID recycling *and* detects zombies
    assert not sleeper.is_running()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import diff_pb2 as diff__pb2


class DiffServiceStub(object):
    """DiffService is a service which provides RPCs to inspect differences
    introduced between a set of commits.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CommitDiff = channel.unary_stream(
                '/gitaly.DiffService/CommitDiff',
                request_serializer=diff__pb2.CommitDiffRequest.SerializeToString,
                response_deserializer=diff__pb2.CommitDiffResponse.FromString,
                )
        self.CommitDelta = channel.unary_stream(
                '/gitaly.DiffService/CommitDelta',
                request_serializer=diff__pb2.CommitDeltaRequest.SerializeToString,
                response_deserializer=diff__pb2.CommitDeltaResponse.FromString,
                )
        self.RawDiff = channel.unary_stream(
                '/gitaly.DiffService/RawDiff',
                request_serializer=diff__pb2.RawDiffRequest.SerializeToString,
                response_deserializer=diff__pb2.RawDiffResponse.FromString,
                )
        self.RawPatch = channel.unary_stream(
                '/gitaly.DiffService/RawPatch',
                request_serializer=diff__pb2.RawPatchRequest.SerializeToString,
                response_deserializer=diff__pb2.RawPatchResponse.FromString,
                )
        self.DiffStats = channel.unary_stream(
                '/gitaly.DiffService/DiffStats',
                request_serializer=diff__pb2.DiffStatsRequest.SerializeToString,
                response_deserializer=diff__pb2.DiffStatsResponse.FromString,
                )
        self.FindChangedPaths = channel.unary_stream(
                '/gitaly.DiffService/FindChangedPaths',
                request_serializer=diff__pb2.FindChangedPathsRequest.SerializeToString,
                response_deserializer=diff__pb2.FindChangedPathsResponse.FromString,
                )
        self.GetPatchID = channel.unary_unary(
                '/gitaly.DiffService/GetPatchID',
                request_serializer=diff__pb2.GetPatchIDRequest.SerializeToString,
                response_deserializer=diff__pb2.GetPatchIDResponse.FromString,
                )
        self.RawRangeDiff = channel.unary_stream(
                '/gitaly.DiffService/RawRangeDiff',
                request_serializer=diff__pb2.RawRangeDiffRequest.SerializeToString,
                response_deserializer=diff__pb2.RawRangeDiffResponse.FromString,
                )
        self.RangeDiff = channel.unary_stream(
                '/gitaly.DiffService/RangeDiff',
                request_serializer=diff__pb2.RangeDiffRequest.SerializeToString,
                response_deserializer=diff__pb2.RangeDiffResponse.FromString,
                )
        self.DiffBlobs = channel.unary_stream(
                '/gitaly.DiffService/DiffBlobs',
                request_serializer=diff__pb2.DiffBlobsRequest.SerializeToString,
                response_deserializer=diff__pb2.DiffBlobsResponse.FromString,
                )


class DiffServiceServicer(object):
    """DiffService is a service which provides RPCs to inspect differences
    introduced between a set of commits.
    """

    def CommitDiff(self, request, context):
        """CommitDiff returns a diff between two different commits. The patch data is
        chunked across messages and get streamed back to the client.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CommitDelta(self, request, context):
        """CommitDelta returns the deltas between two different commits. A delta
        includes everything that changed about a set of paths except for the actual
        diff.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RawDiff(self, request, context):
        """RawDiff returns a diff between two commits. The output is the unmodified
        output from git-diff(1). This is not to be confused with git-diff(1)'s
        --raw mode.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RawPatch(self, request, context):
        """RawPatch returns a diff between two commits in a formatted patch.The output
        is the unmodified output from git-format-patch(1). This is not to be confused with
        git-diff(1)'s --raw mode.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DiffStats(self, request, context):
        """DiffStats returns the diff stats between two commits such as number of lines
        changed, etc.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindChangedPaths(self, request, context):
        """FindChangedPaths returns a list of files changed along with the status of each file
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPatchID(self, request, context):
        """GetPatchID computes a patch ID for a patch. Patch IDs are a unique ID computed by hashing
        a patch with some parameters like line numbers ignored. The patch ID can thus be used to compare
        whether diffs make the same change. Please refer to git-patch-id(1) for further information.
        If the difference between old and new change is empty then this RPC returns an error.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RawRangeDiff(self, request, context):
        """RawRangeDiff outputs the raw range diff data for a given range specification.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RangeDiff(self, request, context):
        """RangeDiff outputs the parsed commit pairs from range diff for a given range specification.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DiffBlobs(self, request, context):
        """DiffBlobs computes diffs between pairs of blobs. A batch of blob pairs is sent to the server.
        The resulting patches are then chucked across response messages and streamed to the client.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DiffServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CommitDiff': grpc.unary_stream_rpc_method_handler(
                    servicer.CommitDiff,
                    request_deserializer=diff__pb2.CommitDiffRequest.FromString,
                    response_serializer=diff__pb2.CommitDiffResponse.SerializeToString,
            ),
            'CommitDelta': grpc.unary_stream_rpc_method_handler(
                    servicer.CommitDelta,
                    request_deserializer=diff__pb2.CommitDeltaRequest.FromString,
                    response_serializer=diff__pb2.CommitDeltaResponse.SerializeToString,
            ),
            'RawDiff': grpc.unary_stream_rpc_method_handler(
                    servicer.RawDiff,
                    request_deserializer=diff__pb2.RawDiffRequest.FromString,
                    response_serializer=diff__pb2.RawDiffResponse.SerializeToString,
            ),
            'RawPatch': grpc.unary_stream_rpc_method_handler(
                    servicer.RawPatch,
                    request_deserializer=diff__pb2.RawPatchRequest.FromString,
                    response_serializer=diff__pb2.RawPatchResponse.SerializeToString,
            ),
            'DiffStats': grpc.unary_stream_rpc_method_handler(
                    servicer.DiffStats,
                    request_deserializer=diff__pb2.DiffStatsRequest.FromString,
                    response_serializer=diff__pb2.DiffStatsResponse.SerializeToString,
            ),
            'FindChangedPaths': grpc.unary_stream_rpc_method_handler(
                    servicer.FindChangedPaths,
                    request_deserializer=diff__pb2.FindChangedPathsRequest.FromString,
                    response_serializer=diff__pb2.FindChangedPathsResponse.SerializeToString,
            ),
            'GetPatchID': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPatchID,
                    request_deserializer=diff__pb2.GetPatchIDRequest.FromString,
                    response_serializer=diff__pb2.GetPatchIDResponse.SerializeToString,
            ),
            'RawRangeDiff': grpc.unary_stream_rpc_method_handler(
                    servicer.RawRangeDiff,
                    request_deserializer=diff__pb2.RawRangeDiffRequest.FromString,
                    response_serializer=diff__pb2.RawRangeDiffResponse.SerializeToString,
            ),
            'RangeDiff': grpc.unary_stream_rpc_method_handler(
                    servicer.RangeDiff,
                    request_deserializer=diff__pb2.RangeDiffRequest.FromString,
                    response_serializer=diff__pb2.RangeDiffResponse.SerializeToString,
            ),
            'DiffBlobs': grpc.unary_stream_rpc_method_handler(
                    servicer.DiffBlobs,
                    request_deserializer=diff__pb2.DiffBlobsRequest.FromString,
                    response_serializer=diff__pb2.DiffBlobsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gitaly.DiffService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class DiffService(object):
    """DiffService is a service which provides RPCs to inspect differences
    introduced between a set of commits.
    """

    @staticmethod
    def CommitDiff(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/CommitDiff',
            diff__pb2.CommitDiffRequest.SerializeToString,
            diff__pb2.CommitDiffResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CommitDelta(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/CommitDelta',
            diff__pb2.CommitDeltaRequest.SerializeToString,
            diff__pb2.CommitDeltaResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RawDiff(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/RawDiff',
            diff__pb2.RawDiffRequest.SerializeToString,
            diff__pb2.RawDiffResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RawPatch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/RawPatch',
            diff__pb2.RawPatchRequest.SerializeToString,
            diff__pb2.RawPatchResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DiffStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/DiffStats',
            diff__pb2.DiffStatsRequest.SerializeToString,
            diff__pb2.DiffStatsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FindChangedPaths(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/FindChangedPaths',
            diff__pb2.FindChangedPathsRequest.SerializeToString,
            diff__pb2.FindChangedPathsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetPatchID(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.DiffService/GetPatchID',
            diff__pb2.GetPatchIDRequest.SerializeToString,
            diff__pb2.GetPatchIDResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RawRangeDiff(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/RawRangeDiff',
            diff__pb2.RawRangeDiffRequest.SerializeToString,
            diff__pb2.RawRangeDiffResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RangeDiff(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/RangeDiff',
            diff__pb2.RangeDiffRequest.SerializeToString,
            diff__pb2.RangeDiffResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DiffBlobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.DiffService/DiffBlobs',
            diff__pb2.DiffBlobsRequest.SerializeToString,
            diff__pb2.DiffBlobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest

from mercurial_testhelpers import (
    as_bytes,
    RepoWrapper,
)
from .. import manifest


@pytest.fixture
def repo_wrapper(tmpdir):
    yield RepoWrapper.init(tmpdir / 'repo')


def test_directory_listings_integration(repo_wrapper):
    wrapper = repo_wrapper
    base_ctx = wrapper.commit_file('foo')

    sub = wrapper.path / 'sub'
    sub.mkdir()
    (sub / 'bar').write('bar content')
    (sub / 'ba2').write('ba2 content')
    deeper = sub / 'deeper'
    deeper.mkdir()
    (deeper / 'ping').write('pong')

    ctx1 = wrapper.commit(rel_paths=['sub/bar', 'sub/ba2', 'sub/deeper/ping'],
                          message="zebar", add_remove=True)

    assert manifest.miner(base_ctx).ls_dir(b'') == ([], [b'foo'])
    # perhaps we'll want an exception there
    assert manifest.miner(base_ctx).ls_dir(b'sub') == ([], [])

    miner1 = manifest.miner(ctx1)
    assert miner1.ls_dir(b'') == ([b'sub'], [b'foo'])
    assert miner1.ls_dir(b'sub') == ([b'sub/deeper'], [b'sub/ba2', b'sub/bar'])
    assert miner1.ls_dir(b'sub/deeper') == ([], [b'sub/deeper/ping'])

    assert list(miner1.iter_dir_recursive(b'')) == [
        (b'foo', False),
        (b'sub', True),
        (b'sub/ba2', False),
        (b'sub/bar', False),
        (b'sub/deeper', True),
        (b'sub/deeper/ping', False)
    ]


def test_iter_dir_with_flat_paths_integration(repo_wrapper):
    wrapper = repo_wrapper
    wrapper.commit_file('foo')

    sub = wrapper.path / 'sub'
    sub.mkdir()
    (sub / 'bar').write('bar content')
    (sub / 'ba2').write('ba2 content')
    deeper = sub / 'deeper'
    deeper.mkdir()
    (deeper / 'ping').write('pong')

    ctx1 = wrapper.commit(rel_paths=['sub/bar', 'sub/ba2', 'sub/deeper/ping'],
                          message="zebar", add_remove=True)

    miner1 = manifest.miner(ctx1)
    assert list(miner1.iter_dir_with_flat_paths(b'')) == [
        (b'foo', False, b'foo'),
        (b'sub', True, b'sub'),
    ]
    assert list(miner1.iter_dir_with_flat_paths(b'sub')) == [
        (b'sub/ba2', False, b'sub/ba2'),
        (b'sub/bar', False, b'sub/bar'),
        (b'sub/deeper', True, b'sub/deeper'),
    ]


class FakeChangeset:
    """Simply encapsulating a list of file paths.

    The goal is to have `self.manifest.iterkeys()` iterate over the
    given file paths.
    """
    def __init__(self, file_paths):
        self.file_paths = [as_bytes(fp) for fp in file_paths]

    def manifest(self):
        return self

    def iterkeys(self):
        return iter(self.file_paths)


def ls_with_flat_paths(path, file_paths):
    """Call iter_dir_with_flat_path for path on a FakeChangeset
    """
    miner = manifest.miner(FakeChangeset(file_paths))
    return list(miner.iter_dir_with_flat_paths(as_bytes(path)))


def reference_ls_flat_paths(top_path, manifest_paths):
    """Reproduction in Python of the reference lang implementation.

    It uses `ManifestMiner.ls_dir()`, which of course would be a
    very inefficient implementation.
    """
    top_path = as_bytes(top_path)
    miner = manifest.miner(FakeChangeset(manifest_paths))

    def ls_dir(path):
        subdirs, filepaths = miner.ls_dir(path)
        entries = [(p, True) for p in subdirs]
        entries.extend((p, False) for p in filepaths)
        entries.sort()
        return entries

    entries = ls_dir(top_path)
    res = []
    # starts to being compared to Gitaly's populateFlatPath here
    # differences:
    # - our entries are just tuples (path, is_dir)
    # - we output a list of (path, is_dir, flat_path)
    # - we don't need a maximum recursion constant in these tests
    for path, is_dir in entries:
        flat_path = path

        if not is_dir:
            res.append((path, is_dir, flat_path))
            continue

        while True:
            sub_entries = ls_dir(flat_path)
            if len(sub_entries) != 1 or not sub_entries[0][1]:
                break
            subentry_dirs, subentry_files = miner.ls_dir(flat_path)
            flat_path = subentry_dirs[0]

        res.append((path, is_dir, flat_path))
    return res


def test_iter_dir_recursive():
    # deeply nested case: no duplicates and in direct traversal order
    miner = manifest.miner(FakeChangeset(('a', 'b/c/d/e', 'b/c/d/f', 'out')))
    assert list(miner.iter_dir_recursive(b'b')) == [
        (b'b/c', True),
        (b'b/c/d', True),
        (b'b/c/d/e', False),
        (b'b/c/d/f', False),
    ]
    assert list(miner.iter_dir_recursive(b'')) == [
        (b'a', False),
        (b'b', True),
        (b'b/c', True),
        (b'b/c/d', True),
        (b'b/c/d/e', False),
        (b'b/c/d/f', False),
        (b'out', False),
    ]


def test_iter_dir_with_flat_path():
    for path, manifest_paths in [
            ('a', ['a/b']),
            ('', ['a/b']),
            ('', ['a/b', 'a/c']),
            ('a', ['a/b/c', 'd']),
            ('a', ['a/b/c', 'a/b/d', 'a/e']),
            ('a', ['a/b/c/d/x', 'a/b/c/d/y', 'a/e']),
            ('a', ['a/b/c/d/x', 'a/b/c/d/y', 'a/b/x']),
            ('a', ['a/b/c', 'a/b/d', 'a/e']),
            ('a', ['a/b/c/d/x', 'a/b/c/d/y', 'a/e']),
            ('a', ['a/b/c/d/x', 'a/b/c/d/y', 'a/b/x']),
    ]:
        actual = ls_with_flat_paths(path, manifest_paths)
        expected = reference_ls_flat_paths(path, manifest_paths)
        assert actual == expected



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import time

import pytest

from ..worker import WorkerProcess


def worker_callable(wid, log_path, waiting_time,
                    *args, **kwargs):  # pragma no cover
    log_path.write("Worker %d starting, remaining args=%r\n" % (wid, args))
    time.sleep(waiting_time)


def ram_eater(wid, log_path, size, **kw):  # pragma no cover
    allocated = list(range(size))
    log_path.write("Worker %d starting, "
                   "allocating a list with %d elements\n" % (wid, size))
    time.sleep(10)
    del allocated


def wait_first_log_line(log_path, timeout=2):
    """Wait for startup to finish by testing log line."""
    start = time.time()
    while time.time() - start < timeout:
        try:
            line = log_path.readlines()[0]
        except Exception:  # pragma no cover
            pass
        else:
            if line.endswith('\n'):
                return line
        time.sleep(timeout / 100)
    return False  # pragma no cover


@pytest.fixture
def worker_log(tmpdir):
    yield tmpdir / 'worker.log'


def test_process_not_created(worker_log):
    wp = WorkerProcess(process_args=(0, worker_log, 0),
                       process_callable=worker_callable)
    assert str(wp) == "Worker 0 (pid: no process)"

    assert not wp.is_alive()

    # methods that should do nothing and certainly not crash
    wp.join()
    wp.terminate()

    # Caller of monitor is not told to stop (next time, the process is
    # expected to have been created)
    assert wp.watch() is True

    # in that case, a restart just starts the process
    wp.restart()
    assert wp.process is not None
    wp.join()

    assert worker_log.read().strip() == (
        "Worker 0 starting, remaining args=()")


def test_restart(worker_log):
    wp = WorkerProcess.run(process_args=(0, worker_log, 10, 'some_arg'),
                           process_callable=worker_callable)
    line = wait_first_log_line(worker_log)
    first_pid = wp.process.pid
    worker_log.remove()  # for clarity (it would be overwritten anyway)

    wp.restart()
    second_pid = wp.process.pid
    assert second_pid != first_pid
    assert wait_first_log_line(worker_log) == line
    wp.terminate()  # let's not wait for 10 useless seconds
    wp.join()

    assert not wp.is_alive()

    # by default, watch() restarts the process if found not running
    assert wp.watch() is True
    third_pid = wp.process.pid
    assert third_pid != second_pid
    assert wait_first_log_line(worker_log) == line
    wp.terminate()
    wp.join()

    # …but not if the appropriate flag is set
    wp.restart_done_workers = False
    assert wp.watch() is False
    fourth_pid = wp.process.pid
    assert fourth_pid == third_pid
    assert not wp.is_alive()


def test_rss_restart(worker_log):
    # we get 53MiB for the RAM eater with size=10000, almost the same with
    # 100000, hence it's mostly Python's overhead. These figures would
    # trigger the restart but perhaps would be too much context-dependent.
    # We finally get and finally 89Mib with size being one million.
    wp = WorkerProcess.run(process_callable=ram_eater,
                           process_args=(0, worker_log, 1000000),
                           max_rss=10 << 20)
    assert wait_first_log_line(worker_log)
    worker_log.remove()

    pid1 = wp.process.pid
    assert wp.watch() is True
    assert wait_first_log_line(worker_log)
    assert wp.process.pid != pid1

    wp.terminate()
    wp.join()



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import logging
import time

from grpc import StatusCode

from heptapod import (
    obsutil,
)

from heptapod.gitlab.branch import (
    NAMED_BRANCH_PREFIX,
)

from ..branch import (
    gitlab_branch_head
)
from ..changelog import (
    ancestor,
    merge_content,
)
from ..errors import (
    operation_error_treatment,
    structured_abort,
)
from ..logging import LoggerAdapter
from ..revision import (
    changeset_by_commit_id_abort,
    gitlab_revision_changeset,
    validate_oid,
)
from ..servicer import HGitalyServicer

from ..stub.operations_pb2 import (
    OperationBranchUpdate,
    UserFFBranchRequest,
    UserFFBranchResponse,
    UserSquashRequest,
    UserSquashResponse,
    UserSquashError,
)
from ..stub.errors_pb2 import (
    ResolveRevisionError,
)
from ..stub.operations_pb2_grpc import OperationServiceServicer


base_logger = logging.getLogger(__name__)


class OperationServicer(OperationServiceServicer, HGitalyServicer):
    """OperationsServiceService implementation.
    """

    def UserSquash(self,
                   request: UserSquashRequest,
                   context) -> UserSquashResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context,
                              for_mutation_by=request.user)
        with_hg_git = not repo.ui.configbool(b'heptapod', b'no-git')
        # Gitaly's squash is merge-centric, start_sha is actually the
        # merge target, whereas end_sha is the head of the MR being accepted
        # TODO check that there are no public changesets for a nicer
        # user feedback.
        start_sha, end_sha = request.start_sha, request.end_sha
        if not start_sha:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty StartSha")
        if not end_sha:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty EndSha")
        start_rev, end_rev = start_sha.encode('ascii'), end_sha.encode('ascii')
        start_ctx = gitlab_revision_changeset(repo, start_rev)
        end_ctx = gitlab_revision_changeset(repo, end_rev)

        for (ctx, rev, error_label) in (
                (start_ctx, start_rev, 'start'), (end_ctx, end_rev, 'end')
        ):
            if ctx is None:
                structured_abort(
                    context,
                    StatusCode.INVALID_ARGUMENT,
                    f'resolving {error_label} revision: reference not found',
                    UserSquashError(resolve_revision=ResolveRevisionError(
                        revision=rev))
                )

        revset = (f"ancestor({start_sha}, {end_sha})::{end_sha}"
                  f"- ancestor({start_sha}, {end_sha})").encode('ascii')
        end_ctx = gitlab_revision_changeset(repo, end_sha.encode('ascii'))
        logger.info("Folding revset %s, mirroring to Git=%r",
                    revset, with_hg_git)
        # TODO add the hg_git flag or maybe let servicer do it.
        message = request.commit_message
        if not message:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty CommitMessage")

        # Mercurial does not distinguish between author (actual author of
        # the work) and committer (could be just someone with commit rights
        # relaying the original work). In case an author is provided, it
        # feels right to derive the Mercurial author from it, as preserving
        # the actual work metadata (and copyright) should have priority.
        # This is what `HgGitRepository` used to do on the Rails side.
        author = request.author
        if not author.name:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty Author")

        # timestamp is supposed to be for the committer, but Mercurial does
        # not have such a distinction, hence it will become the commit date.
        if not request.HasField('timestamp'):
            unix_ts = int(time.time())
        else:
            unix_ts = request.timestamp.seconds

        opts = {'from': False,  # cannot be used as regular kwarg
                'exact': True,
                'rev': [revset],
                'message': message,
                'user': b'%s <%s>' % (author.name, author.email),
                'date': b'%d 0' % unix_ts,
                }
        # Comment taken from `hg_git_repository.rb`:
        #   Note that `hg fold --exact` will fail unless the revset is
        #   "an unbroken linear chain". That fits the idea of a Merge Request
        #   neatly, and should be unsuprising to users: it's natural to expect
        #   squashes to stay simple.
        #   In particular, if there's a merge of a side topic, it will be
        #   unsquashable.
        # Not 100% sure we need a workdir, but I don't see
        #   an explicit "inmemory" option as there is for `hg rebase`. What
        #   I do see is user (status) messages as in updates, so…
        # If we update the workdir to "end" changeset, then the fold will
        #   look like an extra head and be rejected (probably because it is
        #   kept active by being the workdir parent).
        #   On the other hand, the "start" changeset is by design of the
        #   method not to be folded and is probably close enough that we
        #   get a reasonable average efficiency.
        with self.working_dir(gl_repo=request.repository,
                              repo=repo,
                              changeset=start_ctx,
                              context=context) as wd:
            # `allowunstable=no` protects us against all instabilities,
            # in particular against orphaning dependent topics.
            # TODO this setting should probably be set in all mutations
            wd.repo.ui.setconfig(b'experimental.evolution', b'allowunstable',
                                 False)
            with operation_error_treatment(context, UserSquashError,
                                           logger=logger):
                self.repo_command(wd.repo, context, 'fold', **opts)
            self.repo_command(wd.repo, context, 'update', start_ctx.hex())
            # The workdir repo does not have to be reloaded, whereas the
            # main repo would. We just need to regrab the end changeset
            # (now obsolete)
            end_ctx = wd.repo.unfiltered()[end_ctx.rev()]
            folded = obsutil.latest_unique_successor(end_ctx)

        return UserSquashResponse(squash_sha=folded.hex().decode('ascii'))

    def UserFFBranch(self,
                     request: UserFFBranchRequest,
                     context) -> UserFFBranchResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context,
                              for_mutation_by=request.user)
        with_hg_git = not repo.ui.configbool(b'heptapod', b'no-git')

        to_publish = changeset_by_commit_id_abort(
            repo, request.commit_id, context)
        if to_publish is None:
            context.abort(
                StatusCode.INTERNAL,
                f'checking for ancestry: invalid commit: "{request.commit_id}"'
            )

        old_id = request.expected_old_oid
        if old_id and not validate_oid(old_id):
            context.abort(StatusCode.INVALID_ARGUMENT,
                          f'cannot parse commit ID: "{old_id}"')

        if not request.branch:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty branch name")

        if not request.branch.startswith(NAMED_BRANCH_PREFIX):
            context.abort(StatusCode.FAILED_PRECONDITION,
                          "Heptapod fast forwards are currently "
                          "for named branches only (no topics nor bookmarks)")

        current_head = gitlab_branch_head(repo, request.branch)
        if to_publish.branch() != current_head.branch():
            context.abort(StatusCode.FAILED_PRECONDITION,
                          "not a fast-forward (Mercurial branch differs)")

        fail = False
        for cs in merge_content(to_publish, current_head):
            if cs.obsolete():
                fail = True
                fail_msg = "is obsolete"
            if cs.isunstable():
                fail = True
                fail_msg = "is unstable"
        if fail:
            context.abort(StatusCode.FAILED_PRECONDITION,
                          f"not a fast-forward (changeset "
                          f"{cs.hex().decode('ascii')} {fail_msg})")

        if old_id and old_id != current_head.hex().decode('ascii'):
            # We did not need to resolve before this, but now we do because
            # Gitaly has a specific error if resolution fails
            if changeset_by_commit_id_abort(repo, old_id, context) is None:
                context.abort(StatusCode.INVALID_ARGUMENT,
                              "cannot resolve expected old object ID: "
                              "reference not found")
            return UserFFBranchResponse()

        if ancestor(to_publish, current_head) != current_head.rev():
            context.abort(StatusCode.FAILED_PRECONDITION,
                          "not fast forward")

        # TODO use phases.advanceboundary directly? Check cache invalidations
        # carefully. ('phases' command is a pain to call and does lots of
        # unnecessary stuff).
        logger.info("All checks passed, now publishing %r, "
                    "mirroring to Git=%r", to_publish, with_hg_git)
        self.publish(to_publish, context)
        return UserFFBranchResponse(branch_update=OperationBranchUpdate(
            commit_id=to_publish.hex().decode('ascii'),
        ))



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import attr
from contextlib import contextmanager
import os
from pathlib import Path
import subprocess

import grpc

from hgitaly.testing.grpc import wait_server_accepts_connection

HGITALY_SOURCE_ROOT = Path(__file__).parent.parent


@attr.s
class RHGitalyServer:
    home_dir = attr.ib()
    sidecar_address = attr.ib()
    env_overrides = attr.ib(default=None)
    """Optional environment variables overrides.

    If not ``None``, this :class:`dict` will be applied to the process
    environment, after the several settings normally performed :meth:`running`.
    To be used in special cases only.
    """

    termination_timeout = attr.ib(default=1)
    """Time allowed for RHGitaly to shutdown after termination request

    Useful to test that termination actually works. Normally it should be
    very fast, even with the `debug` build.
    """

    socket_name = attr.ib(default='rhgitaly.socket')

    @contextmanager
    def running(self):
        env = dict(os.environ)
        env['GITALY_TESTING_NO_GIT_HOOKS'] = "1"  # will be eventually useful
        timeout = int(env.pop('RHGITALY_STARTUP_TIMEOUT', '30').strip())

        env['RHGITALY_REPOSITORIES_ROOT'] = str(self.home_dir / 'default')
        socket_path = self.home_dir / self.socket_name
        url = 'unix:%s' % socket_path.resolve()
        env['RHGITALY_LISTEN_URL'] = url
        env['RHGITALY_SIDECAR_ADDRESS'] = self.sidecar_address

        if self.env_overrides is not None:
            env.update(self.env_overrides)
        rhgitaly_dir = HGITALY_SOURCE_ROOT / 'rust/rhgitaly'

        rhgitaly_exe = env.get('RHGITALY_EXECUTABLE')
        if rhgitaly_exe is None:  # pragma no cover
            subprocess.check_call(('cargo', 'build', '--locked'),
                                  cwd=rhgitaly_dir)
            run_cmd = ('cargo', 'run')
        else:  # pragma no cover
            # Popen would not run a relative binary so easily
            run_cmd = [Path(rhgitaly_exe).resolve()]

        with open(self.home_dir / 'rhgitaly.log', 'w') as logf:
            rhgitaly = subprocess.Popen(
                run_cmd,
                stdout=logf, stderr=logf,
                env=env,
                cwd=rhgitaly_dir,
            )

        self.pid = rhgitaly.pid
        # visible and useful only by putting a breakpoint right before it:
        print(f"Tests dir: {self.home_dir}")

        try:
            # tonic (actually H2) does not accept the default authority
            # set in the case of Unix Domain Sockets by gRPC 1.58.
            # See https://github.com/hyperium/tonic/issues/742 and for instance
            # this solution by k8s clients:
            # https://github.com/kubernetes/kubernetes/pull/112597
            with grpc.insecure_channel(
                    url,
                    options=[('grpc.default_authority', 'localhost')]
            ) as channel:
                wait_server_accepts_connection(channel, timeout=timeout)
                yield channel
        finally:
            rhgitaly.terminate()
            rhgitaly.wait(timeout=self.termination_timeout)



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import grpc
import logging

from mercurial import (
    error,
)

from .. import message
from ..util import (
    chunked,
)

from ..stub.mercurial_changeset_pb2 import (
    ListMercurialChangesetsRequest,
    ListMercurialChangesetsResponse,
    MercurialChangesetField,
    MercurialRepositoryView,
)
from ..stub.mercurial_changeset_pb2_grpc import (
    MercurialChangesetServiceServicer,
)
from ..servicer import HGitalyServicer

logger = logging.getLogger(__name__)


class MercurialChangesetServicer(MercurialChangesetServiceServicer,
                                 HGitalyServicer):
    """MercurialChangesetService implementation.

    The ordering of methods in this source file is the same as in the proto
    file.
    """
    def ListMercurialChangesets(self,
                                request: ListMercurialChangesetsRequest,
                                context) -> ListMercurialChangesetsResponse:
        repo = self.load_repo(request.repository, context)
        if request.view == MercurialRepositoryView.UNFILTERED:
            repo = repo.unfiltered()

        try:
            revs = repo.revs(request.revset)
        except error.RepoLookupError as exc:
            context.abort(grpc.StatusCode.INVALID_ARGUMENT,
                          "%r (hint=%r)" % (exc, exc.hint))

        fields = set(request.fields)
        if MercurialChangesetField.ALL in fields:
            fields = None

        for chunk in chunked(revs):
            yield ListMercurialChangesetsResponse(
                changesets=(message.mercurial_changeset(repo[rev],
                                                        fields=fields)
                            for rev in chunk)
            )



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from multiprocessing import Process, Pipe
import shutil
import signal

import warnings

from mercurial import (
    error,
    hg,
)
from mercurial.merge import merge
from mercurial.node import nullrev as NULL_REVISION
from hgext3rd.heptapod.branch import set_default_gitlab_branch

import pytest
from pytest_cov.embed import cleanup_on_signal

from heptapod.testhelpers import (
    LocalRepoWrapper,
)

from ..workdir import (
    ROSTER_FILE_NAME,
    WorkingDirectory,
    WorkingDirectoryError,
    locked_roster,
    roster_iter,
    wd_path,
    release_workdir,
    remove_all_workdirs,
    reserve_workdir,
    working_directory,
    workdirs_gc,
)

parametrize = pytest.mark.parametrize


@pytest.fixture
def wd_fixture(tmpdir):
    """Shared fixture for working directories tests

    the first object is the `Path` to the working directories root.
    The second object is the wrapper for a freshly created, empty,
    repository.
    """
    workdirs_root = tmpdir / 'working-directories'
    wrapper = LocalRepoWrapper.init(tmpdir / 'repo',
                                    config=dict(
                                     extensions=dict(topic='', evolve=''),
                                    ))
    yield workdirs_root, wrapper


def test_working_directory_basic(wd_fixture):
    wds_root, wrapper = wd_fixture
    src_repo = wrapper.repo

    with working_directory(wds_root, src_repo) as wd:
        wd_path = wd.path
        wd_id = wd.id
        # caller would typically use `wd.repo`, yet not necessarily
        wd_wrapper = LocalRepoWrapper.load(wd.path)
        sha = wd_wrapper.commit_file('foo').hex()

    # The commit done in the temporary working directory is visbile from
    # the main repository.
    wrapper.reload()

    ctx = src_repo[sha]
    assert ctx.hex() == sha

    # Simple reuse, with an error in usage
    try:
        with working_directory(wds_root, src_repo) as wd:
            assert wd.path == wd_path
            raised = RuntimeError("Workdir tests error")
            raise raised
    except RuntimeError as caught:
        assert caught is raised

    # The working dir has been released despite of the RuntimeError,
    # we can use it again
    with working_directory(wds_root, src_repo, changeset=ctx) as wd:
        assert wd.path == wd_path
        wd_wrapper = LocalRepoWrapper.load(wd.path)
        assert wd_wrapper.repo[sha].branch() == b'default'

    # Two working directories for the same branch
    with working_directory(wds_root, src_repo, changeset=ctx) as wd1:
        with working_directory(wds_root, src_repo, changeset=ctx) as wd2:
            assert wd1.path == wd_path
            assert wd2.path != wd_path
            wd2_path = wd2.path
            wd2_id = wd2.id

    # We now have two available working directories for the default branch
    with src_repo.vfs(ROSTER_FILE_NAME, b'rb') as rosterf:
        wds = {wd_id: branch
               for (wd_id, _, _, branch), _l in roster_iter(rosterf)}
    assert wds == {wd_id: b'default', wd2_id: b'default'}

    # Both can be reused
    with working_directory(wds_root, src_repo, changeset=ctx) as wd1:
        with working_directory(wds_root, src_repo, changeset=ctx) as wd2:
            assert set((wd1.path, wd2.path)) == {wd_path, wd2_path}


def test_working_directory_subrepos(wd_fixture):
    wds_root, wrapper = wd_fixture
    src_repo = wrapper.repo
    nested_path = wrapper.path / 'nested'
    nested = LocalRepoWrapper.init(nested_path)
    nested.write_commit("bar", content="in nested")

    wrapper.path.join('.hgsub').write("nested = nested")
    wrapper.command(b'add', subrepos=True)
    wrapper.command(b'commit', subrepos=True, message=b"invalid")
    changeset = wrapper.repo[b'tip']

    (wrapper.path / '.hg/hgrc').write('\n'.join((
        '[extensions]',
        'heptapod=',
        ''
        )))
    wrapper.reload()

    src_repo = wrapper.repo
    # perhaps one day there will not be an error any more, check
    # heptapod#310 and heptapod#1287
    with pytest.raises(WorkingDirectoryError) as exc_info:
        with working_directory(wds_root, src_repo, changeset=changeset):
            pass
    assert exc_info.value.args[0] == 'update'


@parametrize('stage', ('create', 'load', 'update'))
def test_working_directory_errors(wd_fixture, monkeypatch, stage):
    def raiser(*a, **kw):
        raise error.Abort(b"test-wd")

    wds_root, wrapper = wd_fixture
    src_repo = wrapper.repo
    ctx = wrapper.commit_file('foo')

    if stage == 'load':
        with working_directory(wds_root, src_repo, changeset=ctx):
            pass
        monkeypatch.setattr(hg, 'repository', raiser)
    elif stage == 'create':
        monkeypatch.setattr(hg, 'share', raiser)
    elif stage == 'update':
        monkeypatch.setattr(WorkingDirectory, 'uncaught_update', raiser)

    with pytest.raises(WorkingDirectoryError) as exc_info:
        with working_directory(wds_root, wrapper.repo, changeset=ctx):
            pass
    assert exc_info.value.args[0] == stage
    assert exc_info.value.args[1].args[0] == b'test-wd'


def test_working_directory_branches(wd_fixture):
    wds_root, wrapper = wd_fixture
    src_repo = wrapper.repo

    cs_default = wrapper.commit_file('foo')
    cs_topic = wrapper.commit_file('bar', content="top bar", topic='zetop')
    cs_other = wrapper.commit_file('branch', branch='other', parent=cs_default)

    with working_directory(wds_root, wrapper.repo, changeset=cs_topic) as wd:
        default_wd_id = wd.id

        wctx = wd.repo[None]
        assert wctx.p1() == cs_topic
        assert wctx.branch() == b'default'
        assert wctx.topic() == b'zetop'

        assert (wd.path / 'foo').exists()
        assert (wd.path / 'bar').exists()

    with working_directory(wds_root, wrapper.repo, changeset=cs_other) as wd:
        assert wd.id != default_wd_id
        other_wd_id = wd.id

        wctx = wd.repo[None]
        assert wctx.p1() == cs_other
        assert wctx.branch() == b'other'
        assert not wctx.topic()

        assert (wd.path / 'foo').exists()
        assert (wd.path / 'branch').exists()
        assert not (wd.path / 'bar').exists()

        other_wd_path = wd.path

    with src_repo.vfs(ROSTER_FILE_NAME, b'rb') as rosterf:
        wds = {wd_id: branch
               for (wd_id, _, _, branch), _l in roster_iter(rosterf)}

    assert wds == {default_wd_id: b'default', other_wd_id: b'other'}

    # If the working directory is missing despite being listed, it is not
    # an issue. Could happen with restore from backup (although we should
    # simply not backup the roster file) or be accidental.
    shutil.rmtree(other_wd_path)
    assert not other_wd_path.exists()

    with working_directory(wds_root, wrapper.repo, changeset=cs_other) as wd:
        assert wd.id == other_wd_id
        assert wd.path == other_wd_path

    # Working directories are reused regardless of topic
    with working_directory(wds_root, wrapper.repo, changeset=cs_default) as wd:
        assert wd.id == default_wd_id

        wctx = wd.repo[None]
        assert wctx.p1() == cs_default
        assert wctx.branch() == b'default'
        assert not wctx.topic()

        # testing purge/clean update
        (wd.path / 'bar').write_bytes(b"Some conflicting unversioned content")

    # Even with conflicting unversioned file, a clean working directory is
    # provided on next usage
    with working_directory(wds_root, wrapper.repo, changeset=cs_topic) as wd:
        assert wd.id == default_wd_id

        wctx = wd.repo[None]
        assert wctx.p1() == cs_topic

        assert (wd.path / 'bar').read_bytes() == b"top bar"

        # Acquire Mercurial's wlock, but doesn't release it, as would happen
        # on SIGKILL. It is essential for the test to be meaningful to keep
        # a reference to the lock.
        wlock = wd.repo.wlock()

    # Working directory locks are broken on entering if needed.
    # This would block otherwise, due to the update to the given changeset:
    with working_directory(wds_root, wrapper.repo, changeset=cs_default) as wd:
        assert wd.id == default_wd_id

    # Ironically, there's a warning to use `wlock.release` instead of `del`
    # to release a lock, which is not exactly why we delete it (mostly keeping
    # flake8 happy with this test code).
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        del wlock

    # working directories on obsolete changesets are legitimate
    wrapper.update_bin(cs_topic.node())
    wrapper.amend_file('bar', content="new_bar")

    with working_directory(wds_root, wrapper.repo, changeset=cs_topic) as wd:
        assert wd.id == default_wd_id

        wctx = wd.repo[None]
        parent_cs = wctx.p1()
        assert parent_cs == cs_topic
        assert parent_cs.obsolete()


@parametrize('merge_result', ('conflict', 'mergeable'))
def test_working_directory_merge(wd_fixture, merge_result):
    wds_root, wrapper = wd_fixture
    commit_file = wrapper.commit_file
    conflict = merge_result == 'conflict'

    # a simple merge setup with a conflicting topic and a non-conflicting one
    ctx0 = commit_file('foo')
    cs_default = commit_file('foo', content="default")
    topic_file = 'foo' if conflict else 'bar'
    cs_topic = commit_file(topic_file, parent=ctx0, topic='top', content="top")

    # leaving merge result in a working directory
    with working_directory(wds_root, wrapper.repo, changeset=cs_default) as wd:
        default_wd_id = wd.id
        stats = merge(wd.repo[cs_topic.rev()])
        if conflict:
            assert stats.unresolvedcount > 0

    # re-acquisition cleans up remnants
    with working_directory(wds_root, wrapper.repo, changeset=cs_default) as wd:
        assert wd.id == default_wd_id

        wctx = wd.repo[None]
        assert wctx.p1() == cs_default
        assert wctx.p2().rev() == NULL_REVISION
        # TODO clean status (provide in mercurial_testhelpers)


@pytest.fixture
def purge_fixture(wd_fixture):
    wds_root, wrapper = wd_fixture
    repo = wrapper.repo
    # using a custom default branch to illustrate that 'default' is not
    # hardcoded
    wrapper.commit_file('foo', branch='mydefault')
    set_default_gitlab_branch(repo, b'branch/mydefault')

    return wd_fixture


def test_workdirs_gc_active_default(purge_fixture):
    wds_root, wrapper = purge_fixture
    repo = wrapper.repo

    # no error if there is no working directory
    workdirs_gc(wds_root, repo, 1)

    wd_path(wds_root, 1).mkdir(parents=True)
    wd_path(wds_root, 2).mkdir()
    with locked_roster(repo) as (inf, outf):
        outf.writelines((b"0 12 10000 mydefault\n",
                         b"1 - 10000 other\n",
                         b"2 - 10000 mydefault\n",
                         ))

    workdirs_gc(wds_root, repo, max_age_seconds=100, now=11000)

    # The active work dir for default branch is kept, considered to be enough
    with locked_roster(repo) as (inf, outf):
        lines = inf.readlines()
        assert lines == [b"0 12 10000 mydefault\n"]


def test_workdirs_gc_stale_default(purge_fixture):
    wds_root, wrapper = purge_fixture
    repo = wrapper.repo

    # no error if there is no working directory
    workdirs_gc(wds_root, repo, 1)

    wd_path(wds_root, 0).mkdir(parents=True)
    wd_path(wds_root, 2).mkdir()
    with locked_roster(repo) as (inf, outf):
        outf.writelines((b"0 - 10000 mydefault\n",
                         b"2 - 15000 mydefault\n",
                         ))

    workdirs_gc(wds_root, repo, max_age_seconds=100, now=20000)

    # The active work dir for default branch is kept, considered to be enough
    with locked_roster(repo) as (inf, outf):
        lines = [parsed for parsed, _line in roster_iter(inf)]
        assert len(lines) == 1
        assert lines[0][3] == b'mydefault'


def test_workdirs_gc_no_default(purge_fixture):
    wds_root, wrapper = purge_fixture
    repo = wrapper.repo

    # no error if there is no working directory
    workdirs_gc(wds_root, repo, 1)

    wd_path(wds_root, 0).mkdir(parents=True)
    wd_path(wds_root, 1).mkdir()
    with locked_roster(repo) as (inf, outf):
        outf.writelines((b"0 - 10000 other\n",
                         b"1 - 15000 another\n",
                         ))

    workdirs_gc(wds_root, repo, max_age_seconds=100, now=20000)

    # The active work dir for default branch is kept, considered to be enough
    with locked_roster(repo) as (inf, outf):
        assert len(inf.read()) == 0


def test_workdirs_gc_rmerror(purge_fixture):
    wds_root, wrapper = purge_fixture
    repo = wrapper.repo

    # no error if there is no working directory
    workdirs_gc(wds_root, repo, 1)

    with locked_roster(repo) as (inf, outf):
        outf.writelines((b"0 - 10000 other\n",
                         b"1 - 15000 mydefault\n",
                         ))

    workdirs_gc(wds_root, repo, max_age_seconds=100, now=20000)

    with locked_roster(repo) as (inf, outf):
        wds = {parsed[0]: parsed[3] for parsed, _line in roster_iter(inf)}
        assert wds == {0: b'other', 1: b'mydefault'}


@parametrize('inner_success', ('ok', 'rm-error'))
def test_remove_all_workdirs(purge_fixture, monkeypatch, inner_success):
    wds_root, wrapper = purge_fixture
    repo = wrapper.repo

    # acquiring like ordinary caller should
    with working_directory(wds_root, repo, changeset=repo[0]) as wd:
        assert wd.path.exists()

    with locked_roster(repo) as (inf, outf):
        assert len(inf.readlines()) == 1  # for comparison with after

    if inner_success == 'rm-error':
        def raiser(*args):
            raise RuntimeError(*args)

        monkeypatch.setattr(shutil, 'rmtree', raiser)

    remove_all_workdirs(wds_root, repo)

    if inner_success == 'ok':
        assert not wd.path.exists()
        assert not wds_root.exists()

    with locked_roster(repo) as (inf, outf):
        assert len(inf.read()) == 0


def assert_recv(pipe, expected, timeout=2):
    """Read data from a pipe and assert it is as expected, without blocking.

    Using this rather than just `pipe.recv()` prevents blocking the
    tests upon errors.

    The default timeout is long, but acceptable to humans and CI jobs.
    """
    assert pipe.poll(timeout=timeout)
    assert pipe.recv() == expected


ROSTER_LOCK_TIMEOUT = 0.1
ROSTER_LOCK_ATTEMPTS_DELAY = ROSTER_LOCK_TIMEOUT / 10


def locking_subpro(repo_path, conn):
    """Subprocess taking the roster lock and then taking commands from a pipe.

    The lock is taken at startup after sending the initial message and
    is held until the 'shutdown' message is received
    """
    cleanup_on_signal(signal.SIGTERM)
    cleanup_on_signal(signal.SIGINT)
    from hgitaly import procutil
    procutil.IS_CHILD_PROCESS = True

    repo = LocalRepoWrapper.load(repo_path).repo
    conn.send('started')
    try:
        with locked_roster(repo, timeout=ROSTER_LOCK_TIMEOUT) as (inf, outf):
            conn.send('locked')
            while True:
                msg = conn.recv()
                if msg == 'shutdown':
                    conn.send('bye')
                    conn.close()
                    return
                if isinstance(msg, tuple) and msg[0] == 'write':
                    outf.write(msg[1].encode('utf8'))
                    # atomictempfile does not implement flush
                    conn.send('written')
                if msg == 'read':
                    if inf is None:
                        conn.send(None)
                    else:
                        pos = inf.tell()
                        inf.seek(0)
                        conn.send(inf.read().decode('utf8'))
                        inf.seek(pos)
    except error.LockHeld:
        conn.send('timeout')


@parametrize('lock_attempts', (1, 4))
def test_locked_roster(wd_fixture, lock_attempts):
    wrapper = wd_fixture[1]
    repo_path = wrapper.path

    pipe1, child_pipe1 = Pipe()
    pipe2, child_pipe2 = Pipe()
    pipe3, child_pipe3 = Pipe()
    pipe4, child_pipe4 = Pipe()
    proc1 = Process(target=locking_subpro, args=(repo_path, child_pipe1))
    proc2 = Process(target=locking_subpro, args=(repo_path, child_pipe2))
    proc3 = Process(target=locking_subpro, args=(repo_path, child_pipe3))
    proc4 = Process(target=locking_subpro, args=(repo_path, child_pipe4))
    procs = [proc1, proc2, proc3, proc4]

    try:
        # proc1 starts, write a line, but does not see it in its input file
        # (atomicity)
        proc1.start()
        assert_recv(pipe1, 'started')
        assert_recv(pipe1, 'locked')
        pipe1.send(('write', 'content1'))
        assert_recv(pipe1, 'written')
        pipe1.send('read')
        assert_recv(pipe1, None)

        # proc2 starts but cannot acquire the lock yet
        proc2.start()
        assert_recv(pipe2, 'started')
        assert not pipe2.poll(
            timeout=ROSTER_LOCK_ATTEMPTS_DELAY * lock_attempts)

        # shutting down proc1
        pipe1.send('shutdown')
        assert_recv(pipe1, 'bye')
        proc1.join()

        # now that proc1 has released the lock, proc2 acquires it and sees the
        # write done by proc1
        assert_recv(pipe2, 'locked')
        pipe2.send('read')
        assert_recv(pipe2, 'content1')

        # proc2 overwrites the file, but does not see the change yet in its
        # input stream
        pipe2.send(('write', 'content2'))
        assert_recv(pipe2, 'written')
        pipe2.send('read')
        assert_recv(pipe2, 'content1')

        # proc3 starts, cannot acquire the lock immediately either
        proc3.start()
        assert_recv(pipe3, 'started')
        assert not pipe3.poll(timeout=ROSTER_LOCK_ATTEMPTS_DELAY)

        # after proc2 shutdown, proc3 sees the new content
        pipe2.send('shutdown')
        assert_recv(pipe2, 'bye')
        proc2.join()
        assert_recv(pipe3, 'locked')
        pipe3.send('read')
        assert_recv(pipe3, 'content2')

        # proc4 starts but proc3 does not release the lock in time
        proc4.start()
        assert_recv(pipe4, 'started')
        assert_recv(pipe4, 'timeout')

        pipe3.send('shutdown')
        assert_recv(pipe3, 'bye')
    finally:
        # avoid blocking the test run if there are errors
        for proc in procs:
            if proc.is_alive():
                proc.terminate()
                proc.join()


def test_roster_lock_breaking(wd_fixture):
    wrapper = wd_fixture[1]
    repo_path = wrapper.path

    pipe1, child_pipe1 = Pipe()
    pipe2, child_pipe2 = Pipe()
    proc1 = Process(target=locking_subpro, args=(repo_path, child_pipe1))
    proc2 = Process(target=locking_subpro, args=(repo_path, child_pipe2))
    procs = [proc1, proc2]

    # let's grab the lock from the main tests process, which is not allowed
    # to take it, as it is not one of the HGitaly worker processes. This
    # simulates the case where the PID has been reused: there *is* a process
    # with that pid.
    try:
        with locked_roster(wrapper.repo) as (inf, outf):
            # proc1 takes the lock, ignoring the lock held with an invalid PID
            proc1.start()
            assert_recv(pipe1, 'started')
            assert_recv(pipe1, 'locked')
            pipe1.send(('write', 'content1'))
            assert_recv(pipe1, 'written')

            # of course the lock taken by proc1 blocks proc2
            # Note that exiting normally the `locked_roster` context manager
            # of the main process would release the lock, even if held by
            # proc1, which looks bad, but is irrelevant: in actual operation,
            # roster locks have to be broken if the holding process have died
            # abruptly enough not to have been able to release the lock.
            proc2.start()
            assert_recv(pipe2, 'started')
            assert not pipe2.poll(timeout=ROSTER_LOCK_ATTEMPTS_DELAY)

            # shutting down proc1
            pipe1.send('shutdown')
            assert_recv(pipe1, 'bye')
            proc1.join()

            # now that proc1 has released the lock, proc2 acquires it
            # and sees the write done by proc1
            assert_recv(pipe2, 'locked')
            pipe2.send('read')
            assert_recv(pipe2, 'content1')
    finally:
        # avoid blocking the test run if there are errors
        for proc in procs:
            if proc.is_alive():
                proc.terminate()
                proc.join()


def reserving_subpro(wds_root, repo_path, conn):
    cleanup_on_signal(signal.SIGTERM)
    cleanup_on_signal(signal.SIGINT)
    repo = LocalRepoWrapper.load(repo_path).repo
    from hgitaly import procutil
    conn.send('started')

    while True:
        msg = conn.recv()
        if msg == 'shutdown':
            conn.send('bye')
            conn.close()
            return

        if msg == 'attach':
            procutil.IS_CHILD_PROCESS = True
            conn.send('service-child')
        if msg == 'detach':
            procutil.IS_CHILD_PROCESS = False
            conn.send('mono-process')
        if msg == 'reserve':
            wd = reserve_workdir(wds_root, repo.ui, repo, b'default')
            conn.send(wd.id)
        if msg == 'release':
            release_workdir(repo, wd)
            conn.send('released')


def test_release_workdir_unknown_process(wd_fixture):
    """Reservations by unknown processes have to be ignored."""
    wds_root, wrapper = wd_fixture
    repo_path = wrapper.path

    pipe1, child_pipe1 = Pipe()
    pipe2, child_pipe2 = Pipe()
    pipe3, child_pipe3 = Pipe()

    proc1 = Process(target=reserving_subpro,
                    args=(wds_root, repo_path, child_pipe1))
    proc2 = Process(target=reserving_subpro,
                    args=(wds_root, repo_path, child_pipe2))
    proc3 = Process(target=reserving_subpro,
                    args=(wds_root, repo_path, child_pipe3))
    procs = [proc1, proc2, proc3]
    try:
        # 1st scenario: working directory reserved with a non-existing PID
        #
        # proc1 makes a wd, then dies before it could release it
        # proc2 detects that proc1 is no longer valid, and reuses the wd as
        # if nothing happened. It is likely that the pid has not been reused
        # yet, though we will need to add an inner test to ensure coverage
        # and avoid flakiness
        proc1.start()
        assert_recv(pipe1, 'started')
        pipe1.send('attach')
        assert_recv(pipe1, 'service-child')
        pipe1.send('reserve')
        assert pipe1.poll(timeout=2)
        wd_id = pipe1.recv()
        # shutdown message would *currently* be enough to get the wished stale
        # workdir, but killing the subprocess makes sure the test will
        # not become tautological if there are changes later on
        proc1.kill()
        proc1.join()

        proc2.start()
        assert_recv(pipe2, 'started')
        pipe2.send('attach')
        assert_recv(pipe2, 'service-child')
        pipe2.send('reserve')
        assert_recv(pipe2, wd_id)
        pipe2.send('release')
        pipe2.send('shutdown')
        proc2.join()

        # 2nd scenario: working directory reserved, PID matches an invalid
        # process (it's been reused since then, perhaps even reboot). We
        # use the main tests process PID as a guaranteed existing yet invalid
        # one (with the current logic). PID 1 would be another possible
        # choice if validity rules change.
        repo = wrapper.repo
        wd = reserve_workdir(wds_root, repo.ui, repo, b'default')
        proc3.start()
        assert_recv(pipe3, 'started')
        pipe3.send('attach')
        assert_recv(pipe3, 'service-child')
        pipe3.send('reserve')
        assert_recv(pipe3, wd.id)
        pipe3.send('release')
    finally:
        # avoid blocking the test run if there are errors
        for proc in procs:
            if proc.is_alive():
                proc.terminate()
                proc.join()



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from base64 import (
    b64encode,
    b64decode,
)
from typing import (
    Tuple,
    Union,
)


def concat_chgsid_path(changeset_id, path):
    # '_' is not part of base64 and still matches '\w' in regexps
    return '_'.join((changeset_id, b64encode(path).decode('ascii')))


def split_chgsid_path(oid):
    split = oid.split('_', 1)
    if len(split) == 1:
        return oid, None
    return split[0], b64decode(split[1])


def tree_oid(repo, changeset_id: str, path: bytes):
    return concat_chgsid_path(changeset_id, path)


def blob_oid(repo, changeset_id, path):
    return concat_chgsid_path(changeset_id, path)


def ctx_blob_oid(changectx, path):
    return blob_oid(changectx.repo(), changectx.hex().decode('ascii'), path)


def extract_tree_oid(repo, oid: str) -> Tuple[str, Union[bytes, None]]:
    """"Return changeset id and path for a tree or commit OID.

    Changeset ID is given in hexadecimal form.

    The returned path is ``None`` if and only if ``oid`` is a commit OID.
    """
    return split_chgsid_path(oid)


def extract_blob_oid(repo, oid: str) -> Tuple[str, bytes]:
    """Return changeset ID and path for a blob OID.

    Changeset ID is given in hexadecimal form.

    Examples:

    """
    return split_chgsid_path(oid)



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import attr
import logging
from typing import Any


_missing = object()

CORRELATION_ID_MD_KEY = 'x-gitlab-correlation-id'


def extract_correlation(context):
    """Extract and cache correlation info for a gRPC context."""

    corr_id = getattr(context, 'correlation_id', _missing)
    if corr_id is _missing:
        for md in context.invocation_metadata():
            if md.key == CORRELATION_ID_MD_KEY:
                corr_id = md.value
                break
        else:
            # avoid searching again
            corr_id = None
        context.correlation_id = corr_id
    return corr_id


@attr.define
class LoggerAdapter(logging.LoggerAdapter):
    logger: Any
    grpc_context: Any

    def process(self, msg, kwargs):
        extra = kwargs.get('extra')
        if extra is None:
            kwargs['extra'] = extra = {}

        # set on context by interceptor
        extra['correlation_id'] = extract_correlation(self.grpc_context)
        return msg, kwargs



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os
import pytest

from .. import stream


def test_aggregate_flush_batch():
    in_data = iter([b'AAA', b'BB', b'B', b'C'])
    max_size = 2
    data = stream.aggregate_flush_batch(in_data, len, max_size)
    data = list(data)
    assert data == [[b'AAA'], [b'BB', b'B'], [b'C']]


def test_concat_resplit():
    in_data = iter([b'AAB', b'BCCDD'])
    max_size = 2
    data = stream.concat_resplit(in_data, max_size)
    data = list(data)
    assert data == [b'AA', b'BB', b'CC', b'DD']


def test_env_write_buffer_size(monkeypatch):
    monkeypatch.setitem(os.environ, 'GITALY_STREAMIO_WRITE_BUFFER_SIZE', '12')

    assert stream.env_write_buffer_size() == 12


def test_iter_boolean_lookahead():
    in_data = iter([b'AAB', b'BCCDD'])
    max_size = 2
    itr = stream.concat_resplit(in_data, max_size)
    batched_data = stream.iter_boolean_lookahead(itr)
    for _ in range(3):
        data, eop = next(batched_data)
        assert len(data) == 2
        assert not eop
    data, eop = next(batched_data)
    assert len(data) == 2
    assert eop
    # with empty iterator
    with pytest.raises(StopIteration):
        data = stream.iter_boolean_lookahead(iter([]))
        next(data)


def test_split_batches():
    in_data = b'AABBCCDD'
    max_size = 2
    data = stream.split_batches(in_data, max_size)
    data = list(data)
    assert data == [b'AA', b'BB', b'CC', b'DD']


def test_chunked_limit():
    chunks = ['AA', 'BB', 'CC']
    limit = stream.chunked_limit
    assert list(limit(chunks, 0)) == []
    assert list(limit(chunks, None)) == chunks
    assert list(limit(chunks, 1)) == ['A']
    assert list(limit(chunks, 2)) == ['AA']
    assert list(limit(chunks, 3)) == ['AA', 'B']



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import attr
import logging
import shutil

from hgitaly import feature
from hgitaly.logging import CORRELATION_ID_MD_KEY
from hgitaly.servicer import (
    HG_GIT_MIRRORING_MD_KEY,
    NATIVE_PROJECT_MD_KEY,
    PY_HEPTAPOD_SKIP_HOOKS,
    SKIP_HOOKS_MD_KEY,
)
from hgitaly.stub.shared_pb2 import Repository

from heptapod.testhelpers import (
        LocalRepoWrapper,
)
from heptapod.testhelpers.git import GitRepo
from hgitaly.tests.common import (
    make_empty_repo_uniform,
)
from hgitaly.service.interceptors import base_logger as intercept_logger

from hgitaly.stub.ref_pb2_grpc import (
    RefServiceStub,
)
from hgitaly.stub.ref_pb2 import (
    ListRefsRequest,
)
from hgitaly.stub.shared_pb2 import (
    User,
)

# for coverage of hgitaly.logging
intercept_logger.setLevel(logging.DEBUG)


@attr.s
class ServiceFixture:
    """Fixture base class to absorb some of the boilerplate.

    See test_ref for example usage.

    There is still a lot of duplication between this class and
    :mod:`hgitaly.tests.common` we should probably merge them together
    at some point.

    Interesting attributes not defined as `attr.ib()`:

    - ``grpc_repo``: gRPC :class:`Repository` instance specifying the
       managed repo
    - ``repo_wrapper``: can be used to handle the managed repo directly
    - ``changesets``: a dict mapping convenient names to changeset contexts
      (returned by :attr:`repo_factory`)
    """

    stub_cls = None
    """Class attribute to be provided by concrete classes."""

    grpc_channel = attr.ib()
    server_repos_root = attr.ib()
    with_repo = attr.ib(default=True)
    """If ``True``, create a repository in the default storage upon setup.

    The repository details are then accessible in :attr:`grpc_repo`
    and :attr:`repo_wrapper`. If :attr:`with_repo` is ``False``, these
    attributes are initialised to None
    """
    auto_cleanup = attr.ib(default=False)
    """If ``True``, remove repositories tracked by this class upon tear down.

    For other repositories than the one created if :attr:`with_repo` is
    ``True``, this relies on downstream code adding them to
    attr:`repos_to_cleanup`.

    This can be useful for certain services because the
    ``server_repos_root`` fixture has the module scope, but most
    services won't need it because the path of the repository provided by
    this class is unique.
    """

    feature_flags = attr.ib(factory=list)
    """Feature flags to send to the server.

    Tests can set feature flags by inner mutation, e.g.::

       fixture.feature_flags.append(('my-flag', True))

    As of this writing, they are passed only if the subclasses uses the
    generic `self.rpc`, something like::

      self.rpc('MyRpc', request)
    """

    correlation_id = attr.ib(default=None)
    """Correlation id, to test/exert the logging system
    """

    repo_factory = attr.ib(default=make_empty_repo_uniform)
    """Callable with the same signature as :func:`make_empty_repo_uniform`.
    """

    def __enter__(self):
        self.stub = self.stub_cls(self.grpc_channel)
        if self.with_repo:
            self.repo_wrapper, self.grpc_repo, self.changesets = (
                self.repo_factory(self.server_repos_root))
            self.repos_to_cleanup = [self.repo_wrapper.path]
        else:
            self.repo_wrapper = self.grpc_repo = self.changesets = None
            self.repos_to_cleanup = []
        return self

    def __exit__(self, *exc_args):
        if not self.auto_cleanup:
            return

        for repo_path in self.repos_to_cleanup:
            if repo_path.exists():
                shutil.rmtree(repo_path)

    def additional_repo(self, rel_path, storage_name='default'):
        """Register an additional repository by storage relative path.

        The repository is not created (a future option may be introduced
        to do it).

        :returns: repository absolute path and :class:`Repository` instance
          (gRPC message).
        """
        repo_path = self.repo_path(rel_path, storage_name=storage_name)
        repo_msg = Repository(storage_name=storage_name,
                              relative_path=rel_path)
        self.repos_to_cleanup.append(repo_path)
        return repo_path, repo_msg

    def storage_path(self, storage_name='default'):
        """Utility method to avoid depending too much on actual disk layout.

        This repeats the actual implementation just once.
        """
        return self.server_repos_root / storage_name

    def repo_path(self, rel_path, **kw):
        """Utility method to avoid depending too much on actual disk layout.

        This makes no assumption whether the repo nor the storage actually
        exist.
        """
        return self.storage_path(**kw) / rel_path

    def make_repo_wrapper(self, rel_path, **kw):
        """Utility method to avoid depending too much on actual disk layout.

        The repository is expected to exist.
        """
        return LocalRepoWrapper.load(self.repo_path(rel_path, **kw))

    def rpc(self, method_name, request):
        """Call a method, taking care of metadata."""
        return getattr(self.stub, method_name)(
            request,
            metadata=self.grpc_metadata()
        )

    def grpc_metadata(self):
        """Bake method call metadata.

        Takes care of feature flags only as of this writing, but could also
        be useful with other metadata (correlation id etc.)
        """
        metadata = feature.as_grpc_metadata(self.feature_flags)
        corr_id = self.correlation_id
        if corr_id is not None:
            metadata.append((CORRELATION_ID_MD_KEY, corr_id))
        return metadata

    def git_repo(self):
        return GitRepo(str(self.repo_wrapper.path) + '.git')


class MutationServiceFixture(ServiceFixture):
    """A feature meant for mutating methods.

    It has a default :attr:`user`, generates the appropriate gRPC metadata,
    etc.
    """
    def __enter__(self):
        super(MutationServiceFixture, self).__enter__()
        self.ref_stub = RefServiceStub(self.grpc_channel)
        self.user = User(gl_id='user-37',
                         name=b'Test User',
                         email=b'testuser@heptapod.test',
                         gl_username='testuser')
        self.repo_wrapper.repo.ui.environ[PY_HEPTAPOD_SKIP_HOOKS] = b'yes'
        return self

    def grpc_metadata(self):
        mds = super(MutationServiceFixture, self).grpc_metadata()
        native = getattr(self, 'hg_native', True)
        mds.append((NATIVE_PROJECT_MD_KEY, str(native)))
        with_mirroring = getattr(self, 'with_hg_git_mirroring', None)
        if native and with_mirroring is not None:
            mds.append((HG_GIT_MIRRORING_MD_KEY, str(with_mirroring)))
        mds.append((SKIP_HOOKS_MD_KEY, 'true'))
        return mds

    def list_refs(self, repo=None):
        if repo is None:
            repo = self.grpc_repo

        return {
            r.name: r.target for resp in self.ref_stub.ListRefs(
                ListRefsRequest(repository=repo))
            for r in resp.references
        }



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Complementary tests of the testing harness for coverage.

Can also be useful to help with compatibility issues in hg-git
"""
import pytest

from hgitaly.gitlab_ref import (
    gitlab_special_ref_target,
)


from hgitaly.stub.commit_pb2 import FindCommitRequest
from hgitaly.stub.commit_pb2_grpc import CommitServiceStub

from . import skip_comparison_tests
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip


@pytest.fixture
def rpc_helper(gitaly_comparison):
    """Provide a fully working RpcCHelper.

    The associated method is taken from `test_commit.test_compare_find_commit`,
    but it should not matter so much as the setup with the
    Mercurial and Git repositories.
    """
    yield gitaly_comparison.rpc_helper(
        stub_cls=CommitServiceStub,
        method_name='FindCommit',
        request_cls=FindCommitRequest,
        request_sha_attrs=['revision'],
        response_sha_attrs=['commit.id', 'commit.parent_ids[]'],
        )


def test_hg2git_and_invalidation(rpc_helper):
    comp = rpc_helper.comparison

    # at startup, despite the absence of the hg_git attribute, its
    # invalidation does not fail
    assert 'hg_git' not in comp.__dict__
    comp.invalidate()
    assert 'hg_git' not in comp.__dict__  # still not there

    # Commits are converted to Git and hg2git gives the correspondence
    wrapper = comp.hg_repo_wrapper
    hg_sha0 = wrapper.commit_file('foo', message="Converted to Git").hex()
    git_sha0 = rpc_helper.hg2git(hg_sha0)
    # this is a real Git commit:
    assert comp.git_repo.branches()[b'branch/default'] == dict(
        title=b"Converted to Git",
        sha=git_sha0
    )
    # check a posteriori that earlier absence checks were relevant
    assert 'hg_git' in comp.__dict__

    # hg2git defaults on its input for this use-case:
    assert rpc_helper.hg2git(b'branch/something') == b'branch/something'

    # before invalidation, the helpers don't see new commits
    hg_sha1 = wrapper.commit_file('foo', message="New commit").hex()
    assert rpc_helper.hg2git(hg_sha1) == hg_sha1

    # solved by invalidation
    comp.invalidate()
    git_sha1 = rpc_helper.hg2git(hg_sha1)
    assert git_sha1 != hg_sha1
    assert comp.git_repo.branches()[b'branch/default'] == dict(
        title=b"New commit",
        sha=git_sha1
    )


def test_write_special_ref(gitaly_comparison):
    comp = gitaly_comparison
    wrapper = comp.hg_repo_wrapper
    git_repo = comp.git_repo
    changeset = wrapper.commit_file('foo')
    hg_sha = changeset.hex()
    git_sha = comp.hg_git.map_git_get(hg_sha)

    ref_name = b'pipelines/1'
    ref_path = b'refs/' + ref_name
    comp.write_special_ref(ref_name, hg_sha)
    assert gitlab_special_ref_target(wrapper.repo, ref_path) == changeset
    assert git_repo.all_refs()[ref_path] == git_sha

    with pytest.raises(LookupError):
        comp.write_special_ref(b'will/not/work', b'0123dead' * 5)

    with pytest.raises(LookupError):
        comp.write_special_ref(b'will/not/work', b'not-even-a-sha')



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""High level utilities for Mercurial manifest handling.

The purpose is both to provide exactly what the HGitaly services need
and to abstract away from the actual manifest implementation: we might
at some point switch to some version of Tree Manifest if one appears that
fills our basic needs (namely to be a peer implementation detail, having no
effect on changeset IDs).

The module is meant to harbour several classes, according to the
underlying Mercurial core implementations. The dispatching is done by
the :func:`miner` fatory function.

It is legitimate to optimize each class according to the (stable) properties
of the core manifest object it works on.
"""
import attr
from mercurial.context import filectx
from mercurial.utils.stringutil import binary as is_binary


@attr.s
class ManifestMiner:
    """High level data extraction for basic Mercurial manifest.
    """
    changeset = attr.ib()

    def ls_dir(self, path):
        """List direct directory contents of path at given changeset.

        Anything from inside subdirectories of ``path`` is ignored.

        :param changeset: a :class:`changectx` instance
        :param bytes path: path in the repository of the directory to list.
           Can be empty for the root directory, but not ``b'/'``.
        :returns: a pair ``(subdirs, filepaths)`` of lists, where
          ``subdirs`` contains the sub directories and ``filepaths`` the direct
          file entries within ``path``.
          Both lists are lexicographically sorted.
          All elements are given by their full paths from the root.
        """
        subtrees = set()
        file_paths = []
        prefix = path.rstrip(b'/') + b'/' if path else path
        prefix_len = len(prefix)
        for file_path in self.changeset.manifest().iterkeys():
            if not file_path.startswith(prefix):
                continue
            split = file_path[prefix_len:].split(b'/', 1)
            if len(split) > 1:
                subtrees.add(prefix + split[0])
            else:
                file_paths.append(file_path)
        file_paths.sort()
        return sorted(subtrees), file_paths

    def iter_dir_recursive(self, path):
        """Iterate on recursive directory contents of path in order.

        :returns: yields pairs (path, is_directory)
        """
        prefix = path.rstrip(b'/') + b'/' if path else path
        prefix_len = len(prefix)
        changeset = self.changeset

        in_dir = False
        seen_subdirs = set()

        for file_path in changeset.manifest().iterkeys():
            if not file_path.startswith(prefix):
                if in_dir:
                    break
                continue  # pragma no cover (see coverage#198 and PEP626)

            in_dir = True

            split = file_path[prefix_len:].rsplit(b'/', 1)

            # accumulate ancestor dirs that need to be yielded
            acc = []
            while len(split) > 1:
                subdir = split[0]
                if subdir in seen_subdirs:
                    # if yielded yet, all its ancestors also are
                    break
                acc.append(subdir)
                seen_subdirs.add(subdir)
                split = subdir.rsplit(b'/', 1)

            for subdir in reversed(acc):
                yield (prefix + subdir, True)

            yield (file_path, False)

    def iter_dir_with_flat_paths(self, path):
        """Iterate on directory direct contents with "flat_path" information.

        :returns: yields triplets (full path, is_dir, flat_path) where
                  ``full_path`` is the path of a file or directory from
                  the repo root, ``is_dir`` indicates whether it is a
                  directory and ``flat_path`` is as explained below.

        About ``flat_path``, here is a comment from the current version of
        commit.proto::

          // Relative path of the first subdir that doesn't have only
          // one directory descendant


        Gitaly reference implementation (Golang)::

          func populateFlatPath(ctx context.Context, c catfile.Batch,
                                entries []*gitalypb.TreeEntry) error {
            for _, entry := range entries {
              entry.FlatPath = entry.Path

              if entry.Type != gitalypb.TreeEntry_TREE {
                continue
              }

              for i := 1; i < defaultFlatTreeRecursion; i++ {
                subEntries, err := treeEntries(
                    ctx, c, entry.CommitOid,
                    string(entry.FlatPath), "", false)

                if err != nil {
                  return err
                }

                if (len(subEntries) != 1 ||
                    subEntries[0].Type != gitalypb.TreeEntry_TREE) {
                  break
                }

                entry.FlatPath = subEntries[0].Path
              }
            }

            return nil
          }

        Implementation for the standard Mercurial manifest has of course
        to be very different, since it lists full paths to leaf
        (non-directory) files. In particular, there are no empty directories.

        The interpretation  is that the "flat path" is the longest
        common directory ancestor of all file entries that are inside the
        considered directory entry. Here is a proof (can certainly be much
        simplified in a direct reasoning not involving reductio at absurdum,
        but writing it down was enough to convince ourselves).

        0. Assumptions and conventions:

           + there is no empty directory (guaranteed by the manifest data
             structure in the Mercurial case).
           + by X≤Y, we mean that Y is a path ancestor of X (inclusive)
           + let E be the considered directory entry and Fp its flat path
             as computed by the Golang implementation

        1. Fp is a common directory ancestor of all file entries f such that
           f≤E

           Let G be the greatest common ancestor of Fp and f.
           The crawling performed by the implementation went through G,
           since Fp≤G, so let's see what happened then.
           If f is a direct child of G, then Fp=G. If not, then let h be the
           direct child of G such that f≤h. h cannot be the only directory
           child of g: otherwise we would have Fp≤h, and h would be a greater
           common ancestor of Fp and f. Therefore the crawling stopped at G,
           hence Fp=G.

        2. Let G be the greatest common directory ancestor of all file entries
           f such that f≤E are ancestors of Fp.

           By derinition of "greatest", G≤Fp because Fp is a common directory
           ancestor.
           Let's see what the crawling implementation did at G if G≠Fp:

           Because of the halting condition of the implementation,
           either Fp has a file child and is therefore the greatest common
           directory ancestor, or Fp has at least two distinct directory
           children. Obviously, Only one of those can be an ancestor of G.
           So let h be a child of Fp such that G≤h is false. There is at least
           a file entry k such that k≤h, because directory entries cannot be
           empty. G is not an ancestor of k, contradicting its definition,
           so G must be equal to Fp.

        This implementation relies on manifest to emit paths in sorted manner.
        """
        prefix = path.rstrip(b'/') + b'/' if path else path
        prefix_len = len(prefix)
        changeset = self.changeset

        in_dir = False
        subdir, flat_path = None, ()

        for file_path in changeset.manifest().iterkeys():
            if not file_path.startswith(prefix):
                if in_dir:
                    break
                continue  # pragma no cover (see coverage#198 and PEP626)

            in_dir = True

            split = file_path[prefix_len:].split(b'/')
            if subdir is not None and split[0] != subdir:
                # we are leaving subdir, yield it
                dir_path = prefix + subdir
                yield (dir_path, True,
                       prefix + b'/'.join(flat_path) if flat_path else dir_path
                       )
                subdir, flat_path = None, ()

            if len(split) == 1:
                yield (file_path, False, file_path)
                subdir, flat_path = None, ()
            elif subdir is None:
                subdir, flat_path = split[0], split[:-1]
                continue

            flat_path = [
                segments[0] for segments in zip(flat_path, split)
                if segments[0] == segments[1]
            ]

        if subdir is not None:
            dir_path = prefix + subdir
            yield (dir_path, True,
                   prefix + b'/'.join(flat_path) if flat_path else dir_path
                   )

    def file_names_by_regexp(self, rx, subdir=b''):
        manifest = self.changeset.manifest()
        subdir_prefix = subdir + b'/' if subdir else b''

        for file_path in manifest.iterkeys():
            if not file_path.startswith(subdir_prefix):
                continue

            if rx is not None and rx.search(file_path) is None:
                continue

            yield file_path

    def iter_files_with_content(self, exclude_binary=False):
        manifest = self.changeset.manifest()
        repo = self.changeset.repo().unfiltered()
        for file_path, file_node, flags in manifest.iterentries():
            # filectx has a `isbinary` method, but it would actually
            # read the data once more (not cached on self)
            content = filectx(repo, file_path, fileid=file_node).data()
            if exclude_binary and is_binary(content):
                continue
            yield file_path, content, flags


def miner(changeset):
    """Return an appropriate manifest extractor for the given changeset.

    This factory function abstracts over possible future manifest
    types, for which we might write different implementations
    """
    return ManifestMiner(changeset)



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Various changelog operations

These are utilities involving several changesets, either in the arguments or
in the response.
"""


def ancestor(cs1, cs2):
    """Greatest common ancestor of two changesets."""
    return cs1.repo().revs("ancestor(%s, %s)", cs1, cs2).first()


def merge_content(source, target):
    """Iterator of changeset contexts for the merge of source into target."""
    repo = source.repo()
    for rev in repo.revs("only(%s, %s)", source, target):
        yield repo[rev]
